from __future__ import annotations

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Any, Dict, Iterable, Literal, Mapping, Sequence, Type, TypeVar, Optional
from urllib.parse import (
    parse_qsl,
    quote,
    unquote,
    urlencode,
    urljoin,
    urlsplit,
    urlunsplit,
)

from yggdrasil.io.parameters import anonymize_parameters

__all__ = [
    "URL",
    "URLResource",
    "get_registered_url_resource",
    "register_url_resource",
    "registered_url_schemes",
    "url_resource_class",
]
T = TypeVar("T", bound="URLResource")

_DEFAULT_PORTS = {"http": 80, "https": 443, "ws": 80, "wss": 443}

_SAFE_PATH = "/:@-._~!$&'()*+,;="
_SAFE_QUERY = "-._~!$'()*+,;=:@/?&="
_SAFE_FRAGMENT = "-._~!$&'()*+,;=:@/?"

_NO_PORT = 0

_REGISTRY: Dict[str, Type["URLResource"]] = {}


def _lower_if(value: str) -> str:
    return value.lower() if value else ""


def _strip_trailing_dot(host: str) -> str:
    return host[:-1] if host.endswith(".") else host


def _normalize_path(path: str) -> str:
    if not path:
        return "/"

    elif path == "/":
        return path

    elif not path.startswith("/"):
        path = "/" + os.path.realpath(path).replace("\\", "/")

    return path


def _remove_default_port(scheme: str, host: str, port: int) -> int:
    if not scheme or not host or port <= 0:
        return _NO_PORT
    return _NO_PORT if _DEFAULT_PORTS.get(scheme) == port else port


def _encode_userinfo(userinfo: str) -> str:
    return quote(userinfo, safe=":!$&'()*+,;=") if userinfo else ""


def _encode_path(path: str) -> str:
    return quote(path, safe=_SAFE_PATH + "%")


def _encode_query(query: str) -> str:
    return quote(query, safe=_SAFE_QUERY + "%")


def _encode_fragment(fragment: str) -> str:
    return quote(fragment, safe=_SAFE_FRAGMENT + "%")


def _decode_maybe(value: str, decode: bool) -> str:
    return unquote(value) if decode and value else value


def _s(value: str | None) -> str:
    return value or ""


def _p(value: int | None) -> int:
    return value or _NO_PORT


def _normalize_query(query: str) -> str:
    query = query.lstrip("?")
    if not query:
        return ""
    items = parse_qsl(query, keep_blank_values=True)
    items.sort(key=lambda item: (item[0], item[1]))
    return urlencode(items, doseq=True)


def _parse_port(value: Any) -> int:
    if value in (None, "", 0):
        return _NO_PORT
    if isinstance(value, int):
        return value if value > 0 else _NO_PORT
    text = str(value)
    return int(text) if text.isdigit() and int(text) > 0 else _NO_PORT


def _parse_netloc(netloc: str, *, decode: bool) -> tuple[str, str, int]:
    if not netloc:
        return "", "", _NO_PORT

    userinfo = ""
    hostport = netloc

    if "@" in netloc:
        userinfo, hostport = netloc.rsplit("@", 1)
        userinfo = _decode_maybe(userinfo, decode)

    port = _NO_PORT

    if hostport.startswith("["):
        rb = hostport.find("]")
        if rb == -1:
            return userinfo, hostport, _NO_PORT

        host = hostport[1:rb]
        rest = hostport[rb + 1 :]
        if rest.startswith(":") and len(rest) > 1:
            port = _parse_port(rest[1:])
        return userinfo, host, port

    if ":" in hostport:
        host, port_text = hostport.rsplit(":", 1)
        port = _parse_port(port_text)
        return userinfo, host, port

    return userinfo, hostport, _NO_PORT


def _looks_like_local_path(raw: str) -> bool:
    if not raw:
        return False

    stripped = raw.strip()
    split = urlsplit(stripped)

    if split.scheme:
        if len(split.scheme) == 1 and len(stripped) >= 3 and stripped[1] == ":" and stripped[2] in ("\\", "/"):
            return True
        return False

    return stripped.startswith(("/", "~", "./", "../")) or "/" in stripped or "\\" in stripped


def _pathlike_to_file_parts(raw: str | Path) -> tuple[str, str]:
    path = raw if isinstance(raw, Path) else Path(raw)
    path = path.expanduser().resolve(strict=False)
    return "file", path.as_posix()


def _normalize_components(
    *,
    scheme: str,
    userinfo: str,
    host: str,
    port: int,
    path: str,
    query: str,
    fragment: str,
) -> tuple[str, str, int, str, str, str]:
    scheme_n = _lower_if(scheme)
    host_n = _strip_trailing_dot(_lower_if(host))
    port_n = _remove_default_port(scheme_n, host_n, port)
    path_n = _normalize_path(path)
    query_n = _normalize_query(query)
    fragment_n = fragment.lstrip("#")

    return scheme_n, host_n, port_n, path_n, query_n, fragment_n


@dataclass(frozen=True, slots=True)
class URL:
    scheme: str = ""
    userinfo: str | None = None
    host: str = ""
    port: int | None = None
    path: str = "/"
    query: str | None = None
    fragment: str | None = None

    _str_enc: str | None = field(default=None, init=False, repr=False, compare=False)
    _str_raw: str | None = field(default=None, init=False, repr=False, compare=False)

    @classmethod
    def empty(cls) -> "URL":
        return _EMPTY_URL

    @classmethod
    def parse(
        cls,
        obj: Any,
        *,
        default_scheme: Optional[str] = None,
        decode: bool = False,
        normalize: bool = True,
    ) -> URL:
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, Mapping):
            return cls.parse_dict(obj, decode=decode, normalize=normalize)

        if isinstance(obj, Path):
            scheme, path = _pathlike_to_file_parts(obj)
            return cls.parse_dict(
                {"scheme": default_scheme or scheme, "path": path},
                decode=decode,
                normalize=normalize,
            )

        if obj is None:
            raise ValueError("Cannot parse URL from None")

        return cls.parse_str(
            str(obj),
            default_scheme=default_scheme,
            decode=decode,
            normalize=normalize
        )

    @classmethod
    def parse_str(
        cls,
        raw: str,
        *,
        default_scheme: Optional[str] = None,
        decode: bool = False,
        normalize: bool = True,
    ) -> URL:
        split = urlsplit(raw)
        userinfo, host, port = _parse_netloc(split.netloc, decode=decode)

        scheme = default_scheme or split.scheme
        path = _decode_maybe(split.path, decode)
        query = _decode_maybe(split.query, decode)
        fragment = _decode_maybe(split.fragment, decode)

        if scheme not in ("file", None) and not host and path:
            if "/" in path:
                host, path = path.split("/", 1)
                path = "/" + path.lstrip("/")
            else:
                host = path
                path = "/"

        if normalize:
            scheme, host, port, path, query, fragment = _normalize_components(
                scheme=scheme,
                userinfo=userinfo,
                host=host,
                port=port,
                path=path,
                query=query,
                fragment=fragment,
            )

        return cls(
            scheme=scheme,
            userinfo=userinfo or None,
            host=host,
            port=port or None,
            path=path or "/",
            query=query or None,
            fragment=fragment or None,
        )

    @classmethod
    def parse_dict(
        cls,
        data: Mapping[str, Any],
        *,
        decode: bool = False,
        normalize: bool = True,
    ) -> URL:
        if not data:
            return cls()

        raw = data.get("url") or data.get("raw")
        if raw is not None:
            return cls.parse_str(str(raw), decode=decode, normalize=normalize)

        scheme = str(data.get("scheme") or "")
        path = str(data.get("path") or "")
        query = str(data.get("query") or "")
        fragment = str(data.get("fragment") or "")

        netloc = data.get("netloc")
        if netloc is None:
            netloc = data.get("authority")

        userinfo = ""
        host = ""
        port = _NO_PORT

        if netloc is not None:
            userinfo, host, port = _parse_netloc(str(netloc), decode=decode)

        if data.get("userinfo") is not None:
            userinfo = _decode_maybe(str(data["userinfo"]), decode)

        if data.get("host") is not None:
            host = _decode_maybe(str(data["host"]), decode)

        if "port" in data:
            port = _parse_port(data.get("port"))

        path = _decode_maybe(path, decode)
        query = _decode_maybe(query, decode)
        fragment = _decode_maybe(fragment, decode)

        if normalize:
            scheme, host, port, path, query, fragment = _normalize_components(
                scheme=scheme,
                userinfo=userinfo,
                host=host,
                port=port,
                path=path,
                query=query,
                fragment=fragment,
            )
        else:
            query = query.lstrip("?")
            fragment = fragment.lstrip("#")

        return cls(
            scheme=scheme,
            userinfo=userinfo or None,
            host=host,
            port=port or None,
            path=path or "",
            query=query or None,
            fragment=fragment or None,
        )

    def __str__(self) -> str:
        return self.to_string()

    def __repr__(self):
        return f"URL<{self.to_string()!r}>"

    def __truediv__(self, other: object) -> URL:
        if not isinstance(other, str):
            raise ValueError(f"Cannot join {self} with {type(other)}")
        return self.with_path(other)

    def _replace(self, **changes: Any) -> URL:
        return replace(self, **changes)

    @property
    def user(self) -> str | None:
        ui = self.userinfo
        if ui is None:
            return None
        if ":" in ui:
            return unquote(ui.split(":", 1)[0])
        return unquote(ui)

    @property
    def password(self) -> str | None:
        ui = self.userinfo
        if ui is None or ":" not in ui:
            return None
        return unquote(ui.split(":", 1)[1])

    def with_user_password(self, user: str | None, password: str | None = None) -> URL:
        if user is None and password is None:
            if self.userinfo:
                return self.with_userinfo(None)
            return self

        user_encoded = "" if user is None else quote(str(user), safe="")
        if password is None:
            return self.with_userinfo(user_encoded)

        password_encoded = quote(str(password), safe="")
        return self.with_userinfo(f"{user_encoded}:{password_encoded}")

    @property
    def query_dict(self) -> Mapping[str, tuple[str, ...]]:
        if not self.query:
            return {}

        out: dict[str, list[str]] = {}
        for key, value in parse_qsl(self.query, keep_blank_values=True):
            out.setdefault(key, []).append(value)

        return {key: tuple(values) for key, values in out.items()}

    @property
    def is_absolute(self) -> bool:
        return bool(self.scheme) and bool(self.host)

    @property
    def is_http(self):
        return self.scheme in ("http", "https")

    def to_string(self, *, encode: bool = True) -> str:
        cache_name = "_str_enc" if encode else "_str_raw"
        cached = getattr(self, cache_name)
        if cached is not None:
            return cached

        scheme = self.scheme
        host = self.host
        userinfo = _s(self.userinfo)
        path = self.path
        query = _s(self.query)
        fragment = _s(self.fragment)
        port = _p(self.port)

        if encode:
            userinfo = _encode_userinfo(userinfo)
            path = _encode_path(path)
            query = _encode_query(query)
            fragment = _encode_fragment(fragment)

        netloc = ""
        if host:
            host_text = f"[{host}]" if ":" in host and not host.startswith("[") else host
            netloc = host_text

            if port > 0:
                netloc = f"{netloc}:{port}"

            if userinfo:
                netloc = f"{userinfo}@{netloc}"

        rendered = urlunsplit((scheme, netloc, path, query, fragment))
        object.__setattr__(self, cache_name, rendered)
        return rendered

    @property
    def authority(self) -> str:
        if not self.host:
            return ""

        host_text = f"[{self.host}]" if ":" in self.host and not self.host.startswith("[") else self.host
        authority = host_text

        if _p(self.port) > 0:
            authority = f"{authority}:{self.port}"

        if self.userinfo:
            authority = f"{self.userinfo}@{authority}"

        return authority

    def join(self, ref: str | URL) -> URL:
        base = self.to_string(encode=True)
        target = ref.to_string(encode=True) if isinstance(ref, URL) else ref
        return URL.parse_str(urljoin(base, target), normalize=True)

    def with_scheme(self, scheme: str | None) -> URL:
        scheme_text = _lower_if(_s(scheme))
        port = _remove_default_port(scheme_text, _strip_trailing_dot(_lower_if(self.host)), _p(self.port))
        return self._replace(scheme=scheme_text, port=port or None)

    def with_userinfo(self, userinfo: str | None) -> URL:
        return self._replace(userinfo=userinfo or None)

    def with_host(self, host: str | None) -> URL:
        host_text = _strip_trailing_dot(_lower_if(_s(host)))
        port = _remove_default_port(self.scheme, host_text, _p(self.port))
        return self._replace(host=host_text, port=port or None)

    def with_path(self, path: str | None) -> URL:
        return self._replace(path=_normalize_path(_s(path)))

    def with_query(self, query: str | None) -> URL:
        query_text = _s(query).lstrip("?")
        return self._replace(query=query_text or None)

    def with_fragment(self, fragment: str | None) -> URL:
        fragment_text = _s(fragment).lstrip("#")
        return self._replace(fragment=fragment_text or None)

    def query_items(self, *, keep_blank_values: bool = True) -> tuple[tuple[str, str], ...]:
        if not self.query:
            return ()
        return tuple(parse_qsl(self.query, keep_blank_values=keep_blank_values))

    def query_mapping(self, *, keep_blank_values: bool = True) -> Mapping[str, list[str]]:
        out: dict[str, list[str]] = {}
        for key, value in self.query_items(keep_blank_values=keep_blank_values):
            out.setdefault(key, []).append(value)
        return out

    def add_query_item(self, key: str, value: str | None, replace: bool = True) -> URL:
        if key is None:
            raise ValueError("key cannot be None")

        key_text = str(key)
        value_text = "" if value is None else str(value)

        items = list(self.query_items(keep_blank_values=True))
        if replace:
            items = [(k, v) for k, v in items if k != key_text]

        items.append((key_text, value_text))
        items.sort(key=lambda item: (item[0], item[1]))
        return self.with_query(urlencode(items, doseq=True))

    def with_query_items(
        self,
        items: Mapping[str, str | Sequence[str]] | Iterable[tuple[str, str]],
        *,
        sort_keys: bool = True,
    ) -> URL:
        pairs: list[tuple[str, str]] = []

        if isinstance(items, Mapping):
            for key, value in items.items():
                if isinstance(value, (list, tuple)):
                    pairs.extend((str(key), str(v)) for v in value)
                else:
                    pairs.append((str(key), str(value)))
        else:
            pairs.extend((str(key), str(value)) for key, value in items)

        if sort_keys:
            pairs.sort(key=lambda item: (item[0], item[1]))

        return self.with_query(urlencode(pairs, doseq=True))

    def anonymize(
        self,
        mode: Literal["remove", "redact"] = "remove",
        *,
        sort_keys: bool = True,
    ) -> URL:
        result = self

        if self.query:
            current = self.query_dict
            anonymized = anonymize_parameters(current, mode=mode)
            if anonymized != current:
                result = result.with_query_items(anonymized, sort_keys=sort_keys)

        if self.userinfo:
            result = result.with_userinfo("<redacted>" if mode == "redact" else None)

        return result


class URLResource(ABC):
    @classmethod
    @abstractmethod
    def url_scheme(cls) -> str:
        raise NotImplementedError

    @abstractmethod
    def to_url(self, scheme: str | None = None) -> URL:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_parsed_url(cls: Type[T], url: URL) -> T:
        raise NotImplementedError

    @classmethod
    def from_url(cls: Type[T], url: URL | str | Any) -> T:
        default_scheme = None if cls is URLResource else cls.url_scheme()
        parsed = URL.parse(url, default_scheme=default_scheme)
        scheme = parsed.scheme.strip().lower()

        if cls is URLResource:
            if not scheme:
                raise ValueError(
                    "Cannot dispatch URLResource.from_url() without a scheme. "
                    f"Got url={parsed.to_string(encode=False)!r}"
                )

            impl = _REGISTRY.get(scheme)
            if impl is None:
                registered = ", ".join(registered_url_schemes()) or "(none)"
                raise ValueError(
                    f"No URLResource registered for scheme {scheme!r}. "
                    f"Registered schemes: {registered}"
                )

            return impl.from_parsed_url(parsed)  # type: ignore[return-value]

        return cls.from_parsed_url(parsed)


def get_registered_url_resource(scheme: str) -> Type["URLResource"] | None:
    return _REGISTRY.get((scheme or "").strip().lower())


def registered_url_schemes() -> tuple[str, ...]:
    return tuple(sorted(_REGISTRY))


def url_resource_class(cls: Type[T] | None = None):
    def decorator(resource_cls: Type[T]) -> Type[T]:
        return register_url_resource(resource_cls)

    return decorator if cls is None else decorator(cls)


def register_url_resource(resource_cls: Type[T]) -> Type[T]:
    if not issubclass(resource_cls, URLResource):
        raise TypeError(f"Can only register URLResource subclasses, got {resource_cls!r}")

    scheme = resource_cls.url_scheme().strip().lower()
    if not scheme:
        raise ValueError(f"{resource_cls.__name__}.url_scheme() must return a non-empty scheme")

    _REGISTRY[scheme] = resource_cls
    return resource_cls


@url_resource_class
@dataclass(frozen=True, slots=True)
class FileResource(URLResource):
    path: Path

    @classmethod
    def url_scheme(cls) -> str:
        return "file"

    def to_url(self, scheme: str | None = None) -> URL:
        target_scheme = (scheme or self.url_scheme()).strip().lower()
        if target_scheme != "file":
            raise ValueError(f"{self.__class__.__name__} only supports the 'file' scheme, got {target_scheme!r}")

        normalized = self.path.expanduser().resolve(strict=False).as_posix()

        # Windows drive path must be /C:/... in file URLs
        if len(normalized) >= 2 and normalized[1] == ":":
            normalized = "/" + normalized

        return URL(
            scheme="file",
            path=normalized,
        )

    @classmethod
    def from_parsed_url(cls: Type["FileResource"], url: URL) -> "FileResource":
        raw_path = url.path or ""
        if not raw_path:
            raise ValueError("File URL is missing a path")

        path_str = raw_path

        # file:///C:/x -> Path("C:/x") on Windows-style input
        if len(path_str) >= 3 and path_str[0] == "/" and path_str[2] == ":":
            path_str = path_str[1:]

        return cls(path=Path(path_str).expanduser().resolve(strict=False))

    @classmethod
    def from_path(cls, value: str | Path) -> "FileResource":
        return cls(path=Path(value).expanduser().resolve(strict=False))

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def suffix(self) -> str:
        return self.path.suffix

    @property
    def exists(self) -> bool:
        return self.path.exists()

    def joinpath(self, *parts: str) -> "FileResource":
        return FileResource(self.path.joinpath(*parts).resolve(strict=False))

    def __truediv__(self, other: str) -> "FileResource":
        if not isinstance(other, str):
            raise ValueError(f"Cannot join {self} with {type(other)}")
        return self.joinpath(other)


_EMPTY_URL = URL()



# yggdrasil/pyutils/dynamic_buffer/__init__.py
"""
Spill-to-disk byte buffer with compression and media-type detection.

Public API
----------

.. code-block:: text

    dynamic_buffer/
    ├── _types.py       BytesLike alias
    ├── _config.py      BufferConfig, DEFAULT_CONFIG
    ├── _codec.py       Codec enum  (magic-byte compression detection)
    ├── _media_type.py  MediaType   (magic-byte / text MIME inference)
    └── _buffer.py      BytesIO     (spill-to-disk buffer)

All four public symbols are re-exported from this package so existing
``from yggdrasil.pyutils.dynamic_buffer import BytesIO`` imports keep working
without modification.

Typical usage::

    from yggdrasil.io import BytesIO, BufferConfig

    buf = BytesIO()
    buf.write(payload)
    buf.seek(0)
    print(buf.media_type)   # e.g. MediaType('application/vnd.apache.parquet')
    print(buf.compression)  # e.g. None  or  <Codec.ZSTD: 'zstd'>
    buf.close()
"""

from __future__ import annotations

from .buffer import BytesIO, BufferLike
from .config import BufferConfig, DEFAULT_CONFIG
from .enums import *
from .headers import get_default_user_agent
from .send_config import SendConfig, SendManyConfig
from .types import BytesLike
from .url import URL



"""Dataclass helpers that integrate with Arrow schemas and safe casting."""
from __future__ import annotations

from dataclasses import MISSING, Field, fields, is_dataclass
from inspect import isclass
from typing import TYPE_CHECKING, Any, Mapping, Sequence, Optional

if TYPE_CHECKING:
    import pyarrow as pa

__all__ = [
    "DATACLASS_ARROW_FIELD_CACHE",
    "dataclass_to_arrow_field",
    "get_from_dict",
    "serialize_dataclass_state",
    "restore_dataclass_state",
    "default_value",
]

DATACLASS_ARROW_FIELD_CACHE: dict[type, "pa.Field"] = {}


def dataclass_to_arrow_field(cls_or_instance: Any) -> "pa.Field":
    """Return a cached Arrow Field describing the dataclass type.

    Args:
        cls_or_instance: Dataclass class or instance.

    Returns:
        Arrow field describing the dataclass schema.

    Raises:
        ValueError: If the input is not a dataclass class or instance.
    """
    if not is_dataclass(cls_or_instance):
        raise ValueError(f"{cls_or_instance!r} is not a dataclass or yggdrasil dataclass")

    cls = cls_or_instance if isclass(cls_or_instance) else cls_or_instance.__class__

    existing = DATACLASS_ARROW_FIELD_CACHE.get(cls)
    if existing is not None:
        return existing

    from yggdrasil.arrow.python_arrow import arrow_field_from_hint

    built = arrow_field_from_hint(cls)
    DATACLASS_ARROW_FIELD_CACHE[cls] = built
    return built


def get_from_dict(
    obj: Mapping[str, Any],
    keys: Sequence[str],
    prefix: Optional[str],
) -> Any:
    """Best-effort field lookup with optional prefix support.

    Lookup order for each key:
      1) obj[key]
      2) obj[prefix + key]

    Returns:
      - first non-MISSING value found
      - MISSING if nothing matched
    """
    for key in keys:
        found = obj.get(key, MISSING)
        if found is not MISSING:
            return found

        if prefix:
            found = obj.get(prefix + key, MISSING)
            if found is not MISSING:
                return found

    return MISSING


def default_value(f: Field[Any], with_factory: bool = True) -> Any:
    """Return the effective default value for a dataclass field.

    Returns:
        - f.default when present
        - f.default_factory() when present
        - MISSING otherwise
    """
    if f.default is not MISSING:
        return f.default

    if with_factory and f.default_factory is not MISSING:  # type: ignore[attr-defined]
        return f.default_factory()  # type: ignore[misc]

    return MISSING


def serialize_dataclass_state(obj: Any) -> dict[str, Any]:
    """Serialize constructor state for a dataclass instance.

    Rules:
      - only init=True fields are considered
      - private fields (name starts with "_") are skipped
      - None values are skipped
      - values equal to their effective default are skipped
      - output is a raw payload dict with no version envelope
    """
    payload: dict[str, Any] = {}

    for f in fields(obj):
        if not f.init or f.name.startswith("_"):
            continue

        value = getattr(obj, f.name)

        if value is None:
            continue

        default = default_value(f, with_factory=True)
        if default is not MISSING and value == default:
            continue

        payload[f.name] = value

    return payload


def restore_dataclass_state(obj: Any, state: Any) -> None:
    """Restore dataclass state from a raw payload dict.

    Rules:
      - None is treated as {}
      - unknown keys are ignored
      - missing init=True fields are filled from effective defaults
      - missing required init=True fields raise TypeError
      - non-init fields are reset to their effective defaults when available

    Raises:
        TypeError: If state is not a dict or a required field is missing.
    """
    if state is None:
        payload: dict[str, Any] = {}
    elif isinstance(state, dict):
        payload = state
    else:
        raise TypeError(f"Invalid pickle state for {type(obj).__name__}: {type(state)!r}")

    known_fields = {f.name: f for f in fields(obj)}

    for name, f in known_fields.items():
        if not f.init:
            continue

        if name in payload:
            value = payload[name]
        else:
            value = default_value(f)
            if value is MISSING:
                raise TypeError(
                    f"Cannot restore {type(obj).__name__}: missing required field {name!r}"
                )

        object.__setattr__(obj, name, value)

    for name, f in known_fields.items():
        if f.init:
            continue

        value = default_value(f)
        if value is not MISSING:
            object.__setattr__(obj, name, value)



# yggdrasil.io.request
from __future__ import annotations

import json as json_module
from dataclasses import MISSING, dataclass, field, replace
from typing import TYPE_CHECKING, Any, Callable, Iterator, Literal, Mapping, MutableMapping, Optional

from yggdrasil.arrow.lib import pyarrow as pa
from yggdrasil.dataclasses.dataclass import get_from_dict
from yggdrasil.io import MediaType
from .buffer import BytesIO
from .enums import GZIP, Codec, MimeType
from .headers import PromotedHeaders, normalize_headers, DEFAULT_HOSTNAME
from .url import URL

if TYPE_CHECKING:
    from .response import Response


__all__ = ["PreparedRequest", "REQUEST_ARROW_SCHEMA"]


REQUEST_ARROW_SCHEMA = pa.schema(
    [
        pa.field(
            "request_method",
            pa.string(),
            nullable=False,
            metadata={"comment": "Method (GET, POST, etc.)"},
        ),
        pa.field(
            "request_url_str",
            pa.string(),
            nullable=False,
            metadata={"comment": "Full request URL as string (deterministic)"},
        ),
        pa.field("request_url_scheme", pa.string(), nullable=True, metadata={"comment": "URL scheme (e.g., http, https)"}),
        pa.field("request_url_userinfo", pa.string(), nullable=True, metadata={"comment": "Userinfo from URL authority (e.g., user:pass). Avoid persisting secrets."}),
        pa.field("request_url_host", pa.string(), nullable=True, metadata={"comment": "Host (domain or IP)"}),
        pa.field("request_url_port", pa.int32(), nullable=True, metadata={"comment": "Port number if explicitly specified"}),
        pa.field("request_url_path", pa.string(), nullable=True, metadata={"comment": "Path component of the URL"}),
        pa.field("request_url_query", pa.string(), nullable=True, metadata={"comment": "Raw query string (without leading '?')"}),
        pa.field("request_url_fragment", pa.string(), nullable=True, metadata={"comment": "Fragment identifier (without leading '#')"}),

        pa.field("request_host", pa.string(), nullable=True, metadata={"comment": "Host header"}),
        pa.field("request_user_agent", pa.string(), nullable=True, metadata={"comment": "User-Agent header"}),
        pa.field("request_accept", pa.string(), nullable=True, metadata={"comment": "Accept header"}),
        pa.field("request_accept_encoding", pa.string(), nullable=True, metadata={"comment": "Accept-Encoding header"}),
        pa.field("request_accept_language", pa.string(), nullable=True, metadata={"comment": "Accept-Language header"}),
        pa.field("request_content_type", pa.string(), nullable=True, metadata={"comment": "Content-Type header"}),
        pa.field("request_content_length", pa.int64(), nullable=False, metadata={"comment": "Content-Length header parsed as integer when possible"}),
        pa.field("request_content_encoding", pa.string(), nullable=True, metadata={"comment": "Content-Encoding header"}),
        pa.field("request_transfer_encoding", pa.string(), nullable=True, metadata={"comment": "Transfer-Encoding header"}),

        pa.field(
            "request_headers",
            pa.map_(pa.string(), pa.string()),
            nullable=False,
            metadata={
                "comment": "Request headers excluding promoted common headers",
                "keys_sorted": "false",
            },
        ),
        pa.field(
            "request_tags",
            pa.map_(pa.string(), pa.string()),
            nullable=False,
            metadata={"comment": "Request tags merged with URL query params; explicit tags win on conflict"},
        ),
        pa.field(
            "request_body",
            pa.binary(),
            nullable=True,
            metadata={"comment": "Raw request body bytes"},
        ),
        pa.field(
            "request_body_hash",
            pa.int64(),
            nullable=True,
            metadata={
                "comment": "Signed Int64 XXH3 digest of response_body",
                "algorithm": "xxh3_64",
            },
        ),
        pa.field(
            "request_sent_at",
            pa.timestamp("us", "UTC"),
            nullable=False,
            metadata={"comment": "UTC timestamp when request was dispatched", "unit": "us", "tz": "UTC"},
        ),
        pa.field(
            "request_sent_at_epoch",
            pa.int64(),
            nullable=False,
            metadata={"comment": "UTC epoch timestamp when request was dispatched", "unit": "us", "tz": "UTC"},
        ),
    ],
    metadata={"comment": "Prepared request flattened into deterministic columns for logging/replay."},
)


@dataclass
class PreparedRequest:
    method: str
    url: URL
    headers: MutableMapping[str, str]
    tags: Optional[MutableMapping[str, str]]
    buffer: Optional[BytesIO]
    sent_at_timestamp: int = field(default=0, hash=False, compare=False)

    before_send: Optional[Callable[["PreparedRequest"], "PreparedRequest"]] = field(
        default=None,
        init=False,
        hash=False,
        compare=False,
    )
    prepare_response: Optional[Callable[["Response"], "Response"]] = field(
        default=None,
        init=False,
        hash=False,
        compare=False,
    )

    @classmethod
    def parse(
        cls,
        obj: Any,
        *,
        normalize: bool = True,
        prefix: str = "request_",
    ) -> "PreparedRequest":
        if isinstance(obj, (str, bytes)):
            obj = json_module.loads(obj)

        if isinstance(obj, Mapping):
            return cls.parse_dict(obj, normalize=normalize, prefix=prefix)

        raise ValueError(f"Cannot make {cls.__name__} from {type(obj)}")

    @classmethod
    def parse_dict(
        cls,
        obj: Mapping[str, Any],
        *,
        normalize: bool = True,
        prefix: str = "request_",
    ) -> "PreparedRequest":
        method = cls._parse_method(obj, prefix=prefix)
        url = cls._parse_url(obj, normalize=normalize, prefix=prefix)
        headers = cls._parse_headers(obj, prefix=prefix)
        tags = cls._parse_tags(obj, prefix=prefix)
        buffer = cls._parse_buffer(obj, prefix=prefix)
        sent_at_timestamp = cls._parse_sent_at_timestamp(obj, prefix=prefix)

        return cls(
            method=method,
            url=url,
            headers=headers,
            tags=tags,
            buffer=buffer,
            sent_at_timestamp=sent_at_timestamp,
        )

    @staticmethod
    def _parse_method(obj: Mapping[str, Any], *, prefix: str) -> str:
        method = get_from_dict(obj, keys=("method",), prefix=prefix)
        return "GET" if method is MISSING or method in (None, "") else str(method)

    @classmethod
    def _parse_url(
        cls,
        obj: Mapping[str, Any],
        *,
        normalize: bool,
        prefix: str,
    ) -> URL:
        url_str = get_from_dict(obj, keys=("url_str", "url", "href", "uri", "request_url_str"), prefix=prefix)
        url_struct = get_from_dict(obj, keys=("url", "request_url"), prefix=prefix)

        if url_str is not MISSING and url_str not in (None, ""):
            return URL.parse(url_str, normalize=normalize)

        if isinstance(url_struct, Mapping):
            return URL.parse(
                {
                    "scheme": url_struct.get("scheme") or "",
                    "userinfo": url_struct.get("userinfo") or "",
                    "host": url_struct.get("host") or "",
                    "port": url_struct.get("port") or 0,
                    "path": url_struct.get("path") or "",
                    "query": url_struct.get("query") or "",
                    "fragment": url_struct.get("fragment") or "",
                },
                normalize=normalize,
            )

        scheme = get_from_dict(obj, ("url_scheme",), prefix=prefix)
        userinfo = get_from_dict(obj, ("url_userinfo",), prefix=prefix)
        host = get_from_dict(obj, ("url_host",), prefix=prefix)
        port = get_from_dict(obj, ("url_port",), prefix=prefix)
        path = get_from_dict(obj, ("url_path",), prefix=prefix)
        query = get_from_dict(obj, ("url_query",), prefix=prefix)
        fragment = get_from_dict(obj, ("url_fragment",), prefix=prefix)

        parts = (scheme, userinfo, host, port, path, query, fragment)
        if not any(part is not MISSING for part in parts):
            raise ValueError(
                "PreparedRequest.parse_dict: missing url/url_str/request_url_str or exploded url fields"
            )

        return URL.parse(
            {
                "scheme": "" if scheme in (MISSING, None) else str(scheme),
                "userinfo": "" if userinfo in (MISSING, None) else str(userinfo),
                "host": "" if host in (MISSING, None) else str(host),
                "port": 0 if port is MISSING or port in (None, "") else int(port),
                "path": "" if path in (MISSING, None) else str(path),
                "query": "" if query in (MISSING, None) else str(query),
                "fragment": "" if fragment in (MISSING, None) else str(fragment),
            },
            normalize=normalize,
        )

    @staticmethod
    def _parse_headers(obj: Mapping[str, Any], *, prefix: str) -> MutableMapping[str, str]:
        headers = get_from_dict(obj, keys=("headers",), prefix="request_")

        if isinstance(headers, Mapping):
            parsed = {str(k): str(v) for k, v in headers.items()}
            if parsed:
                return parsed

        dumped_headers = {
            "Host": get_from_dict(obj, keys=("host",), prefix=prefix),
            "User-Agent": get_from_dict(obj, keys=("user_agent",), prefix=prefix),
            "Accept": get_from_dict(obj, keys=("accept",), prefix=prefix),
            "Accept-Encoding": get_from_dict(obj, keys=("accept_encoding",), prefix=prefix),
            "Accept-Language": get_from_dict(obj, keys=("accept_language",), prefix=prefix),
            "Content-Type": get_from_dict(obj, keys=("content_type",), prefix=prefix),
            "Content-Length": get_from_dict(obj, keys=("content_length",), prefix=prefix),
            "Content-Encoding": get_from_dict(obj, keys=("content_encoding",), prefix=prefix),
            "Transfer-Encoding": get_from_dict(obj, keys=("transfer_encoding",), prefix=prefix),
            "Location": get_from_dict(obj, keys=("location",), prefix=prefix),
            "ETag": get_from_dict(obj, keys=("etag",), prefix=prefix),
            "Last-Modified": get_from_dict(obj, keys=("last_modified",), prefix=prefix),
        }

        out: MutableMapping[str, str] = {}
        for header_name, value in dumped_headers.items():
            if value is MISSING or value in (None, ""):
                continue
            out[header_name] = str(value)

        return out

    @staticmethod
    def _parse_tags(obj: Mapping[str, Any], *, prefix: str) -> dict[str, str]:
        tags = get_from_dict(obj, keys=("tags", "request_tags"), prefix=prefix)
        if not isinstance(tags, Mapping):
            return {}
        return {str(k): str(v) for k, v in tags.items()}

    @staticmethod
    def _parse_buffer(obj: Mapping[str, Any], *, prefix: str) -> Optional[BytesIO]:
        buffer = get_from_dict(obj, keys=("buffer", "body", "content", "data"), prefix=prefix)
        if buffer is MISSING or buffer is None:
            return None
        return BytesIO.parse(buffer)

    @staticmethod
    def _parse_sent_at_timestamp(obj: Mapping[str, Any], *, prefix: str) -> int:
        value = get_from_dict(
            obj,
            keys=(
                "sent_at_timestamp",
                "sent_at_timestamp_epoch",
                "sent_at",
                "request_sent_at_epoch",
                "request_sent_at",
            ),
            prefix=prefix,
        )
        if value is MISSING or value in (None, ""):
            return 0
        return int(value)

    def copy(
        self,
        *,
        method: Optional[str] = None,
        url: URL | str | None = None,
        headers: Optional[Mapping[str, str]] = None,
        buffer: Optional[BytesIO] = ...,
        tags: Optional[Mapping[str, str]] = None,
        sent_at_timestamp: Optional[int] = None,
        before_send: Optional[Callable[["PreparedRequest"], "PreparedRequest"]] = ...,
        prepare_response: Optional[Callable[["Response"], "Response"]] = ...,
        normalize: bool = True,
        copy_buffer: bool = False,
    ) -> "PreparedRequest":
        new_url = self.url if url is None else URL.parse(url, normalize=normalize)

        if headers is None:
            new_headers = dict(self.headers) if self.headers else {}
        else:
            new_headers = {str(k): str(v) for k, v in headers.items()}

        if buffer is ...:
            new_buffer = self.buffer
            if copy_buffer and new_buffer is not None:
                new_buffer = BytesIO.parse(new_buffer.to_bytes())
        else:
            new_buffer = buffer

        new_before_send = self.before_send if before_send is ... else before_send
        new_prepare_response = self.prepare_response if prepare_response is ... else prepare_response

        built = self.__class__(
            method=self.method if method is None else str(method),
            url=new_url,
            headers=new_headers,
            buffer=new_buffer,
            tags=self.tags if tags is None else tags,
            sent_at_timestamp=self.sent_at_timestamp if sent_at_timestamp is None else int(sent_at_timestamp),
        )

        built.before_send = new_before_send
        built.prepare_response = new_prepare_response

        return built

    @classmethod
    def prepare(
        cls,
        method: str,
        url: URL | str,
        headers: Optional[MutableMapping[str, str]] = None,
        body: Optional[Any] = None,
        tags: Optional[Mapping[str, str]] = None,
        before_send: Optional[Callable[["PreparedRequest"], "PreparedRequest"]] = None,
        after_received: Optional[Callable[["Response"], "Response"]] = None,
        *,
        json: Optional[Any] = None,
        normalize: bool = True,
        compress_threshold: Optional[int] = 4 * 1024 * 1024,
        compress_codec: Optional[Codec] = GZIP,
    ) -> "PreparedRequest":
        parsed_url = URL.parse(url, normalize=normalize)
        out_headers: MutableMapping[str, str] = {str(k): str(v) for k, v in (headers or {}).items()}

        request_body: Optional[BytesIO] = None
        if body is not None:
            request_body = BytesIO(body, copy=False)
        elif json is not None:
            request_body = BytesIO(json_module.dumps(json).encode("utf-8"), copy=False)
            out_headers["Content-Type"] = MimeType.JSON.value

            if compress_threshold and request_body.size > compress_threshold:
                request_body = request_body.compress(codec=compress_codec)
                out_headers["Content-Encoding"] = compress_codec.name

        if request_body is not None:
            out_headers["Content-Length"] = str(request_body.size)

        built = cls(
            method=str(method),
            url=parsed_url,
            headers=normalize_headers(out_headers, is_request=True, body=request_body) if normalize else out_headers,
            buffer=request_body,
            tags=tags,
            sent_at_timestamp=0,
        )

        built.before_send = before_send
        built.prepare_response = after_received

        return built

    def prepare_to_send(
        self,
        sent_at_timestamp: int,
        headers: Optional[Mapping[str, str]],
    ) -> "PreparedRequest":
        instance = self.before_send(self) if self.before_send else self

        if instance.headers is None:
            instance.headers = {}

        if headers:
            instance.headers.update(headers)

        instance.sent_at_timestamp = sent_at_timestamp

        return instance

    @property
    def body(self) -> Optional[BytesIO]:
        return self.buffer

    @property
    def content_length(self) -> int:
        if self.buffer is not None:
            return self.buffer.size
        return 0

    @property
    def authorization(self) -> Optional[str]:
        return self.headers.get("Authorization") if self.headers else None

    @property
    def accept_media_type(self):
        if not self.headers:
            return MediaType(MimeType.OCTET_STREAM, None)

        accept = MimeType.parse(self.headers.get("Accept"), default=MimeType.OCTET_STREAM)
        codec = Codec.parse(self.headers.get("Accept-Encoding"), default=None)

        return MediaType(accept, codec)

    @accept_media_type.setter
    def accept_media_type(self, value: MediaType):
        if self.headers is None:
            self.headers = {}
        self.headers["Accept"] = value.mime_type.value
        if value.codec:
            self.headers["Accept-Encoding"] = value.codec.name
        else:
            self.headers.pop("Accept-Encoding", None)

    @authorization.setter
    def authorization(self, value: Optional[str]):
        if self.headers is None:
            self.headers = {}
        if value is not None:
            self.headers["Authorization"] = value
        else:
            self.headers.pop("Authorization", None)

    @property
    def x_api_key(self) -> Optional[str]:
        return self.headers.get("X-API-Key") if self.headers else None

    @x_api_key.setter
    def x_api_key(self, value: Optional[str]):
        if self.headers is None:
            self.headers = {}
        if value is not None:
            self.headers["X-API-Key"] = value
        else:
            self.headers.pop("X-API-Key", None)

    def update_headers(
        self,
        headers: Mapping[str, str],
        *,
        normalize: bool = True,
    ) -> "PreparedRequest":
        if not headers:
            return self

        if normalize:
            headers = normalize_headers(
                headers,
                is_request=True, anonymize=False, add_missing=False
            )

        if not self.headers:
            self.headers = headers
            return self

        self.headers.update(headers)

        return self

    def update_tags(
        self,
        tags: MutableMapping[str, str],
    ) -> "PreparedRequest":
        if not tags:
            return self

        if not self.tags:
            self.tags = tags
            return self

        self.tags.update(tags)

        return self

    def anonymize(self, mode: Literal["remove", "redact"] = "remove") -> "PreparedRequest":
        return replace(
            self,
            headers=normalize_headers(
                self.headers,
                is_request=True, mode=mode, body=self.body, anonymize=True
            ),
            url=self.url.anonymize(mode=mode),
        )

    def to_arrow_batch(self, parse: bool = False) -> pa.RecordBatch:
        """Serialise this request into a single-row :class:`pyarrow.RecordBatch`.

        Parameters
        ----------
        parse:
            Reserved for future use.  Currently raises
            :exc:`NotImplementedError` when ``True``.

        Returns
        -------
        pyarrow.RecordBatch
            A single-row batch conforming to :data:`REQUEST_ARROW_SCHEMA`.
        """
        if parse:
            raise NotImplementedError

        u = self.url
        url_s = u.to_string()

        promoted = PromotedHeaders.extract(self.headers or {})

        tags_v = dict(u.query_items())
        if self.tags:
            tags_v.update({str(k): str(v) for k, v in self.tags.items()})

        if self.buffer is not None:
            body_bytes = self.buffer.to_bytes()
            body_hash = self.buffer.xxh3_int64()
        else:
            body_bytes = None
            body_hash = None

        values = {
            "request_method": self.method,
            "request_url_str": url_s,
            "request_url_scheme": u.scheme,
            "request_url_userinfo": u.userinfo,
            "request_url_host": u.host,
            "request_url_port": u.port,
            "request_url_path": u.path,
            "request_url_query": u.query,
            "request_url_fragment": u.fragment,

            "request_host": promoted.host or DEFAULT_HOSTNAME,
            "request_user_agent": promoted.user_agent,
            "request_accept": promoted.accept,
            "request_accept_encoding": promoted.accept_encoding,
            "request_accept_language": promoted.accept_language,
            "request_content_type": promoted.content_type,
            "request_content_length": promoted.content_length or 0,
            "request_content_encoding": promoted.content_encoding,
            "request_transfer_encoding": promoted.transfer_encoding,

            "request_headers": promoted.remaining,
            "request_tags": tags_v,
            "request_body": body_bytes,
            "request_body_hash": body_hash,

            "request_sent_at": self.sent_at_timestamp,
            "request_sent_at_epoch": self.sent_at_timestamp,
        }

        arrays = [
            pa.array([values[f.name]], type=f.type)
            for f in REQUEST_ARROW_SCHEMA
        ]

        return pa.RecordBatch.from_arrays(arrays, schema=REQUEST_ARROW_SCHEMA)  # type: ignore[arg-type]

    def to_arrow_table(self, parse: bool = False) -> pa.Table:
        """Serialise this request into a single-row :class:`pyarrow.Table`.

        Wraps :meth:`to_arrow_batch` in a table.

        Parameters
        ----------
        parse:
            Forwarded to :meth:`to_arrow_batch`.

        Returns
        -------
        pyarrow.Table
        """
        return pa.Table.from_batches([self.to_arrow_batch(parse=parse)])

    # ------------------------------------------------------------------
    # from_arrow
    # ------------------------------------------------------------------

    @classmethod
    def from_arrow(
        cls,
        batch: pa.RecordBatch | pa.Table,
        *,
        normalize: bool = True,
    ) -> "Iterator[PreparedRequest]":
        """Yield :class:`PreparedRequest` objects from an Arrow batch or table.

        Each row in *batch* / *table* is reconstructed into a
        :class:`PreparedRequest` using the same logic that
        :meth:`Response.from_arrow` uses for the request portion of its
        schema.

        Parameters
        ----------
        batch:
            An Arrow :class:`~pyarrow.RecordBatch` or :class:`~pyarrow.Table`
            whose schema is (or is a superset of) :data:`REQUEST_ARROW_SCHEMA`.
        normalize:
            Whether to normalise reconstructed headers and URLs.

        Yields
        ------
        PreparedRequest
            One instance per row, in row order.
        """

        def _iter_batches(obj: pa.RecordBatch | pa.Table) -> Iterator[pa.RecordBatch]:
            if isinstance(obj, pa.RecordBatch):
                yield obj
            else:
                yield from obj.to_batches()

        req_cols = [f.name for f in REQUEST_ARROW_SCHEMA]

        for rb in _iter_batches(batch):
            cols = {
                name: rb.column(name)
                for name in req_cols
                if name in rb.schema.names
            }

            for i in range(rb.num_rows):
                yield cls._from_arrow_cols(cols, i, normalize=normalize)

    @classmethod
    def _from_arrow_cols(
        cls,
        cols: dict[str, Any],
        i: int,
        *,
        normalize: bool = True,
    ) -> "PreparedRequest":
        """Reconstruct a single :class:`PreparedRequest` from Arrow columns.

        Parameters
        ----------
        cols:
            ``{column_name: arrow_column}`` mapping.
        i:
            Row index.
        normalize:
            Whether to normalise headers and URLs.

        Returns
        -------
        PreparedRequest
        """
        def _get(name: str) -> Any:
            if name in cols:
                return cols[name][i].as_py()
            return None

        method = _get("request_method") or "GET"

        url_str = _get("request_url_str")
        if url_str not in (None, ""):
            url_val = str(url_str)
            url_struct = None
        else:
            scheme   = _get("request_url_scheme")
            userinfo = _get("request_url_userinfo")
            host     = _get("request_url_host")
            port     = _get("request_url_port")
            path     = _get("request_url_path")
            query    = _get("request_url_query")
            fragment = _get("request_url_fragment")

            if any(part not in (None, "", 0)
                   for part in (scheme, userinfo, host, port, path, query, fragment)):
                url_val = None
                url_struct = {
                    "scheme":   scheme or "",
                    "userinfo": userinfo or "",
                    "host":     host or "",
                    "port":     0 if port in (None, "") else int(port),
                    "path":     path or "/",
                    "query":    query or "",
                    "fragment": fragment or "",
                }
            else:
                url_val = ""
                url_struct = None

        # Reconstruct headers from promoted columns + remaining map
        _map_val = _get("request_headers")
        try:
            req_headers: dict[str, str] = (
                {str(k): str(v) for k, v in _map_val if k is not None and v is not None}
                if _map_val else {}
            )
        except Exception:
            req_headers = {}

        for hk, col_name in (
            ("Host",              "request_host"),
            ("User-Agent",        "request_user_agent"),
            ("Accept",            "request_accept"),
            ("Accept-Encoding",   "request_accept_encoding"),
            ("Accept-Language",   "request_accept_language"),
            ("Content-Type",      "request_content_type"),
            ("Content-Length",    "request_content_length"),
            ("Content-Encoding",  "request_content_encoding"),
            ("Transfer-Encoding", "request_transfer_encoding"),
        ):
            hv = _get(col_name)
            if hv not in (None, ""):
                req_headers[hk] = str(hv)

        # Tags
        _tags_val = _get("request_tags")
        try:
            tags: dict[str, str] = (
                {str(k): str(v) for k, v in _tags_val if k is not None and v is not None}
                if _tags_val else {}
            )
        except Exception:
            tags = {}

        # Timestamp
        sent_at_col = cols.get("request_sent_at")
        if sent_at_col is not None:
            ts_raw = sent_at_col[i].as_py()
            if ts_raw is not None and hasattr(ts_raw, "timestamp"):
                sent_at = int(ts_raw.timestamp() * 1_000_000)
            elif isinstance(ts_raw, (int, float)):
                sent_at = int(ts_raw)
            else:
                sent_at = 0
        else:
            sent_at = 0

        return cls.parse_dict(
            {
                "method":            method,
                "url_str":           url_val,
                "url":               url_struct,
                "headers":           req_headers,
                "tags":              tags,
                "buffer":            _get("request_body"),
                "sent_at_timestamp": sent_at,
            },
            normalize=normalize,
        )

    def apply(
        self,
        func: Callable[["PreparedRequest"], "PreparedRequest"],
    ):
        return func(self)


"""
Cluster management helpers for Databricks compute.

This module provides a lightweight ``Cluster`` helper that wraps the
Databricks SDK to simplify common CRUD operations and metadata handling
for clusters. Metadata is stored in custom tags prefixed with
``yggdrasil:``.
"""

from __future__ import annotations

import dataclasses
import inspect
import logging
import re
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Union, TypeVar, \
    TYPE_CHECKING

from databricks.sdk import ClustersAPI
from databricks.sdk.client_types import ClientType
from databricks.sdk.errors import DatabricksError
from databricks.sdk.errors.platform import ResourceDoesNotExist
from databricks.sdk.service._internal import Wait
from databricks.sdk.service.compute import (
    ClusterAccessControlRequest,
    ClusterDetails,
    ClusterPermissionLevel,
    Language,
    Library,
    LibraryInstallStatus,
    PythonPyPiLibrary,
    State, )

from yggdrasil.dataclasses.waiting import WaitingConfig, WaitingConfigArg
from yggdrasil.environ import PyEnv, UserInfo
from yggdrasil.environ.pip_settings import PipIndexSettings
from yggdrasil.io.url import URL
from yggdrasil.version import VersionInfo
from .execution_context import ExecutionContext
from .service import Clusters, PYTHON_BY_DBR
from ..client import DatabricksResource
from ...pyutils.equality import dicts_equal

if TYPE_CHECKING:
    from .command_execution import CommandExecution


__all__ = ["Cluster"]


LOGGER = logging.getLogger(__name__)
F = TypeVar("F", bound=Callable[..., Any])
_EDIT_ARG_NAMES = set(inspect.signature(ClustersAPI.edit).parameters.keys())

_CLUSTER_RUNTIME_FIELDS = frozenset({"_system_context", "_contexts"})
_CLUSTER_SKIP_IF_NONE = frozenset({"_details", "cluster_name"})


def _library_sig(lib: Library) -> tuple:
    # stable-ish signature for dedupe
    if getattr(lib, "jar", None):
        return "jar", lib.jar
    if getattr(lib, "whl", None):
        return "whl", lib.whl
    if getattr(lib, "requirements", None):
        return "req", lib.requirements
    if getattr(lib, "pypi", None) and lib.pypi:
        return "pypi", lib.pypi.package, lib.pypi.repo
    return "other", repr(lib)


@dataclass
class Cluster(DatabricksResource):
    service: Clusters = dataclasses.field(
        default_factory=Clusters.current,
        repr=False,
        compare=False
    )
    cluster_id: Optional[str] = None
    cluster_name: Optional[str] = None

    _details: Optional[ClusterDetails] = dataclasses.field(default=None, repr=False, hash=False, compare=False)
    _details_refresh_time: float = dataclasses.field(default=0.0, repr=False, hash=False, compare=False)
    _contexts: dict[str, ExecutionContext] = dataclasses.field(default_factory=dict, repr=False, hash=False, compare=False)

    def __post_init__(self):
        super().__post_init__()

        if self.cluster_name and not self.cluster_id:
            found = self.service.find_cluster(
                cluster_name=self.cluster_name, raise_error=True
            )

            object.__setattr__(self, "cluster_id", found.cluster_id)
            object.__setattr__(self, "_details", found._details)

    def __repr__(self):
        return "%s(url=%s)" % (self.__class__.__name__, self.url())

    def __str__(self):
        return self.url().to_string()

    def url(self) -> URL:
        return URL.parse_str("%s/compute/clusters/%s" % (
            self.client.base_url.to_string().rstrip("/"),
            self.cluster_id or "unknown"
        ))

    # ------------------------------------------------------------------ #
    # Details caching
    # ------------------------------------------------------------------ #
    @property
    def details(self) -> Optional[ClusterDetails]:
        if self._details is None and self.cluster_id is not None:
            self.set_details(self.clusters_client().get(cluster_id=self.cluster_id))
        return self._details

    def fresh_details(self, max_delay: float | None = None) -> Optional[ClusterDetails]:
        max_delay = 0.0 if max_delay is None else float(max_delay)
        delay = time.time() - float(self._details_refresh_time)

        if self.cluster_id and delay > max_delay:
            self.set_details(self.clusters_client().get(cluster_id=self.cluster_id))
        return self._details

    def refresh(self, max_delay: float | None = None):
        self.set_details(self.fresh_details(max_delay=max_delay))
        return self

    def set_details(self, details: Optional[ClusterDetails]):
        if isinstance(details, ClusterDetails):
            object.__setattr__(self, "_details_refresh_time", time.time())
            object.__setattr__(self, "_details", details)
            object.__setattr__(self, "cluster_id", details.cluster_id)
            object.__setattr__(self, "cluster_name", details.cluster_name)
        elif isinstance(details, Wait):
            # allow passing Wait from start/restart_and_wait without forcing an extra get call
            object.__setattr__(self, "_details_refresh_time", time.time())
            object.__setattr__(self, "_details", details.result())
            object.__setattr__(self, "cluster_id", details.cluster_id)

            if self.cluster_id and details.cluster_id != self.cluster_id:
                object.__setattr__(self, "cluster_name", None)

        else:
            object.__setattr__(self, "_details_refresh_time", 0.0)
            object.__setattr__(self, "_details", None)

        return self

    @property
    def state(self):
        self.refresh()
        if self._details is not None:
            return self._details.state
        return State.UNKNOWN

    @property
    def is_running(self):
        return self.state == State.RUNNING

    @property
    def is_pending(self):
        return self.state in (State.PENDING, State.RESIZING, State.RESTARTING, State.TERMINATING)

    @property
    def is_error(self):
        return self.state == State.ERROR

    @property
    def requirements(self):
        return self.context(context_key="system").requirements

    def raise_for_status(self):
        if self.is_error:
            raise DatabricksError("Error in %s" % self)
        return self

    def wait_for_status(self, wait: WaitingConfigArg = True, raise_error: bool = True):
        wait = WaitingConfig.check_arg(wait)
        if wait:
            iteration, start = 0, time.time()
            while self.is_pending:
                wait.sleep(iteration=iteration, start=start)
                iteration += 1

            self.wait_installed_libraries(wait=wait, raise_error=raise_error)

            if raise_error:
                self.raise_for_status()

        return self

    # ------------------------------------------------------------------ #
    # Runtime helpers
    # ------------------------------------------------------------------ #
    @property
    def spark_version(self) -> Optional[str]:
        d = self.details
        return None if d is None else d.spark_version

    @property
    def runtime_version(self) -> Optional[str]:
        v = self.spark_version
        if not v:
            return None
        parts = v.split(".")
        if len(parts) < 2:
            return None
        return ".".join(parts[:2])

    @property
    def python_version_info(self) -> Optional[VersionInfo]:
        v = self.runtime_version
        if not v:
            return None
        return PYTHON_BY_DBR.get(v)

    # ------------------------------------------------------------------ #
    # Internal helpers
    # ------------------------------------------------------------------ #
    def clusters_client(self) -> ClustersAPI:
        return self.client.workspace_client().clusters

    def update(
        self,
        *,
        single_user_name: Optional[str] = None,
        libraries: Optional[list[Union[str, Library]]] = None,
        permissions: Optional[list[str | ClusterAccessControlRequest]] = None,
        wait: WaitingConfigArg = True,
        **cluster_spec: Any,
    ) -> "Cluster":
        self.install_libraries(libraries=libraries, wait=False, raise_error=False)

        existing_details = {k: v for k, v in self.details.as_shallow_dict().items() if k in _EDIT_ARG_NAMES}

        update_details = {
            k: v for k, v in self.service.check_details(
                update=True, details=self.details,
                single_user_name=single_user_name,
                **cluster_spec
            ).as_shallow_dict().items() if k in _EDIT_ARG_NAMES
        }

        same = dicts_equal(existing_details, update_details, keys=_EDIT_ARG_NAMES)

        if not same:
            LOGGER.debug("Updating %s with %s", self, update_details)

            self.wait_for_status(wait=wait)
            self.clusters_client().edit(**update_details)
            self.update_permissions(permissions=permissions)

            LOGGER.info("Updated %s", self)
            self.wait_for_status(wait=wait)

        return self

    def update_permissions(
        self,
        permissions: Optional[list[str | ClusterAccessControlRequest]] = None
    ):
        if not permissions:
            return self

        permissions = self._check_permission(permissions)

        try:
            self.clusters_client().update_permissions(cluster_id=self.cluster_id, access_control_list=permissions)
        except ResourceDoesNotExist as e:
            _GROUPNAME_RE = re.compile(r"\bGroupName\((?P<group>[^)]*)\)")
            m = _GROUPNAME_RE.search(str(e))
            group_name = m.group("group") if m else None

            if group_name:
                try:
                    self.client.iam.groups.create(
                        name=group_name,
                        members=[self.client.iam.users.current_user],
                        client_type=ClientType.ACCOUNT
                    )
                except Exception as inner_e:
                    raise inner_e from e
                return self.update_permissions(permissions)
            else:
                raise

        return self

    def _check_permission(
        self,
        permission: Union[str, ClusterAccessControlRequest, list[Union[str, ClusterAccessControlRequest]]],
    ):
        if isinstance(permission, ClusterAccessControlRequest):
            return permission

        if isinstance(permission, str):
            if "@" in permission:
                group_name, user_name = None, permission
            else:
                group_name, user_name = permission, None

            return ClusterAccessControlRequest(
                group_name=group_name,
                user_name=user_name,
                permission_level=ClusterPermissionLevel.CAN_MANAGE,
            )

        return [self._check_permission(_) for _ in permission]

    def ensure_running(self, wait: WaitingConfigArg = True) -> "Cluster":
        return self.start(wait=wait)

    def start(self, wait: WaitingConfigArg = True) -> "Cluster":
        if self.is_running:
            return self

        client = self.clusters_client()
        wait = WaitingConfig.check_arg(wait)

        LOGGER.debug("Starting %s", self)

        try:
            client.start(cluster_id=self.cluster_id)
        except DatabricksError:
            self.wait_for_status(wait=wait)
            if self.is_running:
                return self
            client.start(cluster_id=self.cluster_id)

        LOGGER.info("Started %s", self)
        self.wait_for_status(wait=wait)
        return self

    def restart(self, wait: WaitingConfigArg = True):
        self.wait_for_status()

        if self.is_running:
            self.set_details(self.clusters_client().restart_and_wait(cluster_id=self.cluster_id))
            return self

        return self.start(wait=wait)

    def delete(self):
        if self.cluster_id:
            LOGGER.debug("Deleting %s", self)
            self.clusters_client().delete(cluster_id=self.cluster_id)
            LOGGER.info("Deleted %s", self)

    # ------------------------------------------------------------------ #
    # Execution contexts
    # ------------------------------------------------------------------ #
    def context(
        self,
        *,
        language: Optional[Language] = None,
        context_id: Optional[str] = None,
        context_key: Optional[str] = None,
    ) -> "ExecutionContext":
        if context_key:
            existing = self._contexts.get(context_key)
            if existing is None:
                existing = self._contexts[context_key] = ExecutionContext(
                    cluster=self,
                    language=language,
                    context_id=context_id,
                    context_key=context_key,
                )
            return existing

        return ExecutionContext(
            cluster=self,
            language=language,
            context_id=context_id,
            context_key=context_key,
        )

    def command(
        self,
        command: Optional[str | Callable] = None,
        *,
        command_str: Optional[str] = None,
        func: Optional[Callable] = None,
        language: Optional[Language] = None,
        command_id: Optional[str] = None,
        environ: Optional[Union[Iterable[str], Dict[str, str]]] = None,
        context_key: Optional[str] = None,
    ) -> "CommandExecution":
        language = Language.PYTHON if language is None else language

        if not context_key:
            usr, env = UserInfo.current(), PyEnv.current()
            vinfo = env.version_info
            context_key = f"{usr.hostname}-py{vinfo.major}.{vinfo.minor}"

        context = self.context(language=language, context_key=context_key)

        return context.command(
            command=command,
            command_str=command_str,
            command_id=command_id,
            environ=environ,
            func=func,
            language=language,
            context=context
        )

    # ------------------------------------------------------------------ #
    # Libraries
    # ------------------------------------------------------------------ #
    def install_libraries(
        self,
        libraries: Optional[List[Union[str, Library]]] = None,
        wait: WaitingConfigArg = True,
        pip_settings: Optional[PipIndexSettings] = None,
        raise_error: bool = True,
    ) -> "Cluster":
        if not libraries:
            return self

        wsdk = self.client.workspace_client()

        pip_settings = PipIndexSettings.current() if pip_settings is None else pip_settings

        normalized: list[Library] = [self._check_library(_, pip_settings=pip_settings) for _ in libraries if _]

        if normalized:
            existing_sigs = {
                _library_sig(st.library)
                for st in self.installed_library_statuses()
                if getattr(st, "library", None) is not None
            }
            normalized = [lib for lib in normalized if _library_sig(lib) not in existing_sigs]

        if normalized:
            wsdk.libraries.install(cluster_id=self.cluster_id, libraries=normalized)
            self.wait_installed_libraries(wait=wait, raise_error=raise_error)

        return self

    def installed_library_statuses(self):
        return self.client.workspace_client().libraries.cluster_status(cluster_id=self.cluster_id)

    def uninstall_libraries(
        self,
        pypi_packages: Optional[list[str]] = None,
        libraries: Optional[list[Library]] = None,
        restart: bool = True,
    ):
        if libraries is None:
            to_remove = [
                lib.library
                for lib in self.installed_library_statuses()
                if self._filter_lib(lib, pypi_packages=pypi_packages, default_filter=False)
            ]
        else:
            to_remove = libraries

        if to_remove:
            self.client.workspace_client().libraries.uninstall(cluster_id=self.cluster_id, libraries=to_remove)

            if restart:
                self.restart()

        return self

    @staticmethod
    def _filter_lib(lib: Optional[Library], pypi_packages: Optional[list[str]] = None, default_filter: bool = False):
        if lib is None:
            return False

        if lib.pypi:
            if lib.pypi.package and pypi_packages:
                return lib.pypi.package in pypi_packages

        return default_filter

    def wait_installed_libraries(
        self,
        wait: WaitingConfigArg | None = True,
        raise_error: bool = True,
    ):
        if not self.is_running:
            return self

        wait = WaitingConfig.check_arg(wait)

        if wait:
            statuses = list(self.installed_library_statuses())
            start, iteration = time.time(), 0

            while True:
                failed = [_.library for _ in statuses if _.library and _.status == LibraryInstallStatus.FAILED]

                if failed:
                    if raise_error:
                        raise DatabricksError("Libraries %s in %s failed to install" % (failed, self))
                    LOGGER.exception("Libraries %s in %s failed to install", failed, self)

                running = [
                    _
                    for _ in statuses
                    if _.status
                    in (LibraryInstallStatus.INSTALLING, LibraryInstallStatus.PENDING, LibraryInstallStatus.RESOLVING)
                ]

                if not running:
                    break

                wait.sleep(iteration=iteration, start=start)
                iteration += 1
                statuses = list(self.installed_library_statuses())

        return self

    @staticmethod
    def _check_library(value, pip_settings: Optional[PipIndexSettings] = None) -> Library:
        if isinstance(value, Library):
            return value

        pip_settings = PipIndexSettings.current() if pip_settings is None else pip_settings

        if isinstance(value, str):
            # Now value is either a dbfs:/ path or plain package name
            if value.endswith(".jar"):
                return Library(jar=value)
            if value.endswith("requirements.txt"):
                return Library(requirements=value)
            if value.endswith(".whl"):
                return Library(whl=value)

            repo = None

            if pip_settings.extra_index_url and (
                value.startswith("datamanagement")
                or value.startswith("TSSecrets")
                or value.startswith("TSMails")
                or value.startswith("tgp_")
                or value.startswith("wma-data")
            ):
                repo = pip_settings.extra_index_url

            return Library(
                pypi=PythonPyPiLibrary(
                    package=value,
                    repo=repo,
                )
            )

        raise ValueError(f"Cannot build Library object from {type(value)}")


from __future__ import annotations

from .builders import _country, _validate_unique_attrs
from .constants import _SRC_GOOGLE_COUNTRIES_CSV
from .geozone import GeoZone

__all__ = ["load_countries"]

COUNTRIES = [
    ("FRANCE", "FR", "France", 46.2276, 2.2137, {"eic": "10YFR-RTE------C", "aliases": ("BZN|FR", "FRANCE_FR"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SWITZERLAND", "CH", "Switzerland", 46.8182, 8.2275, {"eic": "10YCH-SWISSGRIDZ", "aliases": ("BZN|CH", "SWITZERLAND_CH"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("UNITED_KINGDOM", "GB", "United Kingdom", 55.3781, -3.4360, {"eic": "10Y1001A1001A92E", "aliases": ("UK", "GBR", "BZN|GB", "UNITED_KINGDOM_GB"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("AUSTRIA", "AT", "Austria", 47.5162, 14.5501, {"eic": "10YAT-APG------L", "aliases": ("BZN|AT",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("BELGIUM", "BE", "Belgium", 50.5039, 4.4699, {"eic": "10YBE----------2", "aliases": ("BZN|BE",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("BULGARIA", "BG", "Bulgaria", 42.7339, 25.4858, {"eic": "10YCA-BULGARIA-R", "aliases": ("BZN|BG",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("CROATIA", "HR", "Croatia", 45.1000, 15.2000, {"eic": "10YHR-HEP------M", "aliases": ("BZN|HR",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("CZECHIA", "CZ", "Czechia", 49.8175, 15.4730, {"eic": "10YCZ-CEPS-----N", "aliases": ("CZ", "CZECH_REPUBLIC", "BZN|CZ"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("DENMARK", "DK", "Denmark", 56.2639, 9.5018, {"eic": "10Y1001A1001A65H", "aliases": ("BZN|DK",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ESTONIA", "EE", "Estonia", 58.5953, 25.0136, {"eic": "10Y1001A1001A39I", "aliases": ("BZN|EE",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("FINLAND", "FI", "Finland", 61.9241, 25.7482, {"eic": "10YFI-1--------U", "aliases": ("BZN|FI",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("GERMANY", "DE", "Germany", 51.1657, 10.4515, {"eic": "10Y1001A1001A83F", "aliases": ("GERMANY_DE", "IPA|DE"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("GREECE", "GR", "Greece", 39.0742, 21.8243, {"eic": "10YGR-HTSO-----Y", "aliases": ("BZN|GR",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("HUNGARY", "HU", "Hungary", 47.1625, 19.5033, {"eic": "10YHU-MAVIR----U", "aliases": ("BZN|HU",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ITALY", "IT", "Italy", 41.8719, 12.5674, {"eic": "10YIT-GRTN-----B", "aliases": ("BZN|IT",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("LATVIA", "LV", "Latvia", 56.8796, 24.6032, {"eic": "10YLV-1001A00074", "aliases": ("BZN|LV",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("LITHUANIA", "LT", "Lithuania", 55.1694, 23.8813, {"eic": "10YLT-1001A0008Q", "aliases": ("BZN|LT",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("LUXEMBOURG", "LU", "Luxembourg", 49.8153, 6.1296, {"eic": "10YLU-CEGEDEL-NQ", "aliases": ("BZN|LU", "LUX"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("MONTENEGRO", "ME", "Montenegro", 42.7087, 19.3744, {"eic": "10YCS-CG-TSO---S", "aliases": ("BZN|ME",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("NETHERLANDS", "NL", "Netherlands", 52.1326, 5.2913, {"eic": "10YNL----------L", "aliases": ("BZN|NL", "HOLLAND"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("NORWAY", "NO", "Norway", 60.4720, 8.4689, {"eic": "10YNO-0--------C", "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("POLAND", "PL", "Poland", 51.9194, 19.1451, {"eic": "10YPL-AREA-----S", "aliases": ("BZN|PL",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("PORTUGAL", "PT", "Portugal", 39.3999, -8.2245, {"eic": "10YPT-REN------W", "aliases": ("BZN|PT",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ROMANIA", "RO", "Romania", 45.9432, 24.9668, {"eic": "10YRO-TEL------P", "aliases": ("BZN|RO",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SERBIA", "RS", "Serbia", 44.0165, 21.0059, {"eic": "10YCS-SERBIATSOV", "aliases": ("BZN|RS",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SLOVAKIA", "SK", "Slovakia", 48.6690, 19.6990, {"eic": "10YSK-SEPS-----K", "aliases": ("BZN|SK",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SLOVENIA", "SI", "Slovenia", 46.1512, 14.9955, {"eic": "10YSI-ELES-----O", "aliases": ("BZN|SI",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SPAIN", "ES", "Spain", 40.4637, -3.7492, {"eic": "10YES-REE------0", "aliases": ("BZN|ES",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SWEDEN", "SE", "Sweden", 60.1282, 18.6435, {"eic": "10YSE-1--------K", "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("UKRAINE", "UA", "Ukraine", 48.3794, 31.1656, {"eic": "10Y1001C--00003F", "aliases": ("BZN|UA",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("KOSOVO", "XK", "Kosovo", 42.6026, 20.9030, {"eic": "10Y1001C--00100H", "aliases": ("BZN|XK",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("MOLDOVA", "MD", "Moldova", 47.4116, 28.3699, {"eic": "10Y1001A1001A990", "aliases": ("BZN|MD",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("MALTA", "MT", "Malta", 35.9375, 14.3754, {"eic": "10Y1001A1001A93C", "aliases": ("BZN|MT",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("IRELAND", "IE", "Ireland", 53.4, -7.7, {"eic": "10Y1001A1001A59C", "aliases": ("IE_SEM", "SEM", "BZN|IE_SEM"), "coord_source": "seed: module (rounded/approx; consider aligning to Google countries.csv)", "confidence": "medium"}),
    ("CYPRUS", "CY", "Cyprus", 35.1264, 33.4299, {"eic": "10YCY-1001A0003J", "aliases": ("BZN|CY",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ALBANIA", "AL", "Albania", 41.1533, 20.1683, {"eic": "10YAL-KESH-----5", "aliases": ("BZN|AL",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("BOSNIA", "BA", "Bosnia and Herzegovina", 43.9159, 17.6791, {"eic": "10YBA-JPCC-----D", "aliases": ("BZN|BA", "BOSNIA_HERZEGOVINA"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("NORTH_MACEDONIA", "MK", "North Macedonia", 41.6086, 21.7453, {"eic": "10YMK-MEPSO----8", "aliases": ("BZN|MK", "MACEDONIA"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("BELARUS", "BY", "Belarus", 53.7098, 27.9534, {"eic": "10Y1001A1001A51S", "aliases": ("BYELORUSSIA",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("GEORGIA", "GE", "Georgia", 42.3154, 43.3569, {"aliases": ("BZN|GE",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ARMENIA", "AM", "Armenia", 40.0691, 45.0382, {"aliases": ("BZN|AM",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("AZERBAIJAN", "AZ", "Azerbaijan", 40.1431, 47.5769, {"aliases": ("BZN|AZ",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("TURKEY", "TR", "Turkey", 38.9637, 35.2433, {"eic": "10YTR-TEIAS----W", "aliases": ("BZN|TR", "TURKIYE"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("RUSSIA", "RU", "Russia", 61.5240, 105.3188, {"aliases": ("RUS",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ICELAND", "IS", "Iceland", 64.9631, -19.0208, {"eic": None, "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("LIECHTENSTEIN", "LI", "Liechtenstein", 47.1660, 9.5554, {"coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("UNITED_STATES", "US", "United States", 39.8283, -98.5795, {"aliases": ("USA",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("CANADA", "CA", "Canada", 56.1304, -106.3468, {"aliases": ("CAN",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("MEXICO", "MX", "Mexico", 23.6345, -102.5528, {"aliases": ("MEX",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("BRAZIL", "BR", "Brazil", -14.2350, -51.9253, {"aliases": ("BRA",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("ARGENTINA", "AR", "Argentina", -38.4161, -63.6167, {"aliases": ("ARG",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("CHILE", "CL", "Chile", -35.6751, -71.5430, {"aliases": ("CHL",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("COLOMBIA", "CO", "Colombia", 4.5709, -74.2973, {"aliases": ("COL",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("JAPAN", "JP", "Japan", 36.2048, 138.2529, {"aliases": ("JPN",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("CHINA", "CN", "China", 35.8617, 104.1954, {"aliases": ("CHN",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("INDIA", "IN", "India", 20.5937, 78.9629, {"aliases": ("IND",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SOUTH_KOREA", "KR", "South Korea", 35.9078, 127.7669, {"aliases": ("KOR", "KOREA"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("AUSTRALIA", "AU", "Australia", -25.2744, 133.7751, {"aliases": ("AUS",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("NEW_ZEALAND", "NZ", "New Zealand", -40.9006, 174.8860, {"aliases": ("NZL",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SINGAPORE", "SG", "Singapore", 1.3521, 103.8198, {"aliases": ("SGP",), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SAUDI_ARABIA", "SA", "Saudi Arabia", 23.8859, 45.0792, {"aliases": ("SAU", "KSA"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("UAE", "AE", "United Arab Emirates", 23.4241, 53.8478, {"aliases": ("AED", "EMIRATES"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
    ("SOUTH_AFRICA", "ZA", "South Africa", -30.5595, 22.9375, {"aliases": ("ZAF", "RSA"), "coord_source": _SRC_GOOGLE_COUNTRIES_CSV}),
]


def load_countries() -> None:
    _validate_unique_attrs(COUNTRIES, "COUNTRIES")
    for attr, iso, name, lat, lon, kwargs in COUNTRIES:
        setattr(GeoZone, attr, _country(iso, name, lat, lon, **kwargs))


"""Databricks integrations and helpers for Yggdrasil."""

from .lib import WorkspaceClient # noqa: F401
from .workspaces import Workspace
from .client import DatabricksClient


__all__ = [
    "DatabricksClient",
    "Workspace",
]



"""Arrow IPC (Feather v2) I/O on top of :class:`~yggdrasil.io.buffer.BytesIO`.

Reads both the *file* and *stream* IPC layouts (tries file first, falls
back to stream).  Writes use the file layout with
:class:`pyarrow.ipc.IpcWriteOptions` for intra-file body compression.

Transport-level compression is handled transparently by the base class.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Self

from .media_io import MediaIO
from .media_options import MediaOptions

if TYPE_CHECKING:
    import pyarrow

__all__ = ["IPCIO", "IPCOptions"]


@dataclass(slots=True)
class IPCOptions(MediaOptions):
    """Options for Arrow IPC I/O.

    Parameters
    ----------
    columns:
        Column names to read (``None`` reads all).
    use_threads:
        Enable multi-threaded reading.
    compression:
        Intra-file body compression codec used by
        :class:`pyarrow.ipc.IpcWriteOptions`.
    """

    columns: list[str] | None = None
    use_threads: bool = True

    compression: str | None = "zstd"

    @classmethod
    def resolve(cls, *, options: Self | None = None, **overrides) -> Self:
        """Merge *overrides* into *options* (or a fresh default)."""
        base = options or cls()
        valid = cls.__dataclass_fields__.keys()  # type: ignore[attr-defined]
        unknown = set(overrides) - set(valid)
        if unknown:
            raise TypeError(f"{cls.__name__}.resolve(): unknown option(s): {sorted(unknown)}")
        for k, v in overrides.items():
            setattr(base, k, v)
        return base


@dataclass(slots=True)
class IPCIO(MediaIO[IPCOptions]):
    """Arrow IPC I/O backed by :mod:`pyarrow.ipc`."""

    @classmethod
    def check_options(cls, options: Optional[IPCOptions], *args, **kwargs) -> IPCOptions:
        """Validate and merge caller-supplied options."""
        return IPCOptions.check_parameters(options=options, **kwargs)

    def _read_arrow_table(self, *, options: IPCOptions) -> "pyarrow.Table":
        """Read IPC bytes from the (uncompressed) buffer.

        Tries the *file* layout first; falls back to *stream* on
        :class:`ArrowInvalid`.
        """
        import pyarrow as pa
        import pyarrow.ipc as ipc

        if self.buffer.size <= 0:
            return pa.Table.from_batches([], schema=pa.schema([]))

        arrow_io = self.buffer.to_arrow_io("r")
        try:
            try:
                reader = ipc.open_file(arrow_io)
            except (pa.ArrowInvalid, pa.ArrowIOError):
                arrow_io.seek(0)
                reader = ipc.open_stream(arrow_io)

            table = reader.read_all()

            if options.columns is not None:
                table = table.select(options.columns)

            return table
        finally:
            arrow_io.close()

    def _write_arrow_table(self, *, table: "pyarrow.Table", options: IPCOptions) -> None:
        """Write an Arrow table as IPC into the (uncompressed) buffer."""
        import pyarrow.ipc as ipc

        arrow_io = self.buffer.to_arrow_io("w")
        try:
            write_options = ipc.IpcWriteOptions(
                compression=options.compression,
                use_legacy_format=False,
                use_threads=True,
                allow_64bit=True,
            )

            with ipc.new_file(
                arrow_io,
                table.schema,
                options=write_options,
            ) as writer:
                writer.write_table(table)
        finally:
            arrow_io.close()


from __future__ import annotations

from dataclasses import dataclass
from typing import Mapping

from yggdrasil.io import BytesIO
from yggdrasil.io.buffer.bytes_view import BytesIOView
from yggdrasil.pickle.ser.constants import HEADER_SIZE
from yggdrasil.pickle.ser.errors import HeaderDecodeError, MetadataDecodeError

__all__ = [
    "Metadata",
    "decode_metadata",
    "encode_metadata",
    "Header",
]

Metadata = dict[bytes, bytes] | None


def encode_metadata(metadata: Mapping[bytes, bytes] | None) -> bytes:
    """Encode metadata as length-prefixed key/value pairs.

    Format:
        [k_len:u32][key][v_len:u32][value]...
    """
    if not metadata:
        return b""

    out = bytearray()
    for key, value in metadata.items():
        if not isinstance(key, bytes):
            raise TypeError(f"Metadata keys must be bytes, got {type(key)!r}")
        if not isinstance(value, bytes):
            raise TypeError(f"Metadata values must be bytes, got {type(value)!r}")

        out += len(key).to_bytes(4, "big")
        out += key
        out += len(value).to_bytes(4, "big")
        out += value
    return bytes(out)


def decode_metadata(blob: bytes) -> Metadata:
    """Decode length-prefixed metadata bytes into a dict."""
    if not blob:
        return None

    pos = 0
    size = len(blob)
    out: dict[bytes, bytes] = {}

    while pos < size:
        if pos + 4 > size:
            raise MetadataDecodeError("Unexpected EOF while reading metadata key length")
        k_len = int.from_bytes(blob[pos : pos + 4], "big")
        pos += 4

        if pos + k_len > size:
            raise MetadataDecodeError("Unexpected EOF while reading metadata key")
        key = blob[pos : pos + k_len]
        pos += k_len

        if pos + 4 > size:
            raise MetadataDecodeError("Unexpected EOF while reading metadata value length")
        v_len = int.from_bytes(blob[pos : pos + 4], "big")
        pos += 4

        if pos + v_len > size:
            raise MetadataDecodeError("Unexpected EOF while reading metadata value")
        value = blob[pos : pos + v_len]
        pos += v_len

        out[key] = value

    return out


@dataclass(frozen=True, slots=True)
class Header:
    """Binary header for a serialized payload.

    Layout:
        tag:u16
        codec:u16
        size:u32
        meta_size:u32
        metadata:meta_size bytes
        payload:size bytes
    """

    tag: int
    codec: int
    size: int
    meta_size: int
    start: int
    metadata: Metadata = None

    def __post_init__(self) -> None:
        if self.tag < 0:
            raise ValueError("tag must be >= 0")
        if self.codec < 0:
            raise ValueError("codec must be >= 0")
        if self.size < 0:
            raise ValueError("size must be >= 0")
        if self.meta_size < 0:
            raise ValueError("meta_size must be >= 0")
        if self.start < 0:
            raise ValueError("start must be >= 0")

    @property
    def header_start(self) -> int:
        return self.start - HEADER_SIZE - self.meta_size

    @property
    def payload_end(self) -> int:
        return self.start + self.size

    @classmethod
    def build(
        cls,
        *,
        tag: int,
        codec: int,
        size: int,
        metadata: Mapping[bytes, bytes] | None = None,
        start: int = 0,
    ) -> "Header":
        """Build a header from semantic fields."""
        encoded = encode_metadata(metadata)
        return cls(
            tag=tag,
            codec=codec,
            size=size,
            meta_size=len(encoded),
            start=start + HEADER_SIZE + len(encoded),
            metadata=dict(metadata) if metadata else None,
        )

    @classmethod
    def read_from(
        cls,
        buffer: BytesIO,
        *,
        pos: int | None = None,
    ) -> "Header":
        """Parse a header from a buffer without materializing the payload."""
        if pos is None:
            pos = buffer.tell()

        fixed = buffer.pread(HEADER_SIZE, pos=pos)
        if len(fixed) != HEADER_SIZE:
            raise HeaderDecodeError(
                f"Expected {HEADER_SIZE} header bytes at pos={pos}, got {len(fixed)}"
            )

        tag = int.from_bytes(fixed[0:2], "big")
        codec = int.from_bytes(fixed[2:4], "big")
        size = int.from_bytes(fixed[4:8], "big")
        meta_size = int.from_bytes(fixed[8:12], "big")

        meta_blob = b""
        if meta_size:
            meta_blob = buffer.pread(meta_size, pos=pos + HEADER_SIZE)
            if len(meta_blob) != meta_size:
                raise HeaderDecodeError(
                    f"Expected {meta_size} metadata bytes at pos={pos + HEADER_SIZE}, "
                    f"got {len(meta_blob)}"
                )

        metadata = decode_metadata(meta_blob)
        return cls(
            tag=tag,
            codec=codec,
            size=size,
            meta_size=meta_size,
            start=pos + HEADER_SIZE + meta_size,
            metadata=metadata,
        )

    def payload_view(self, buffer: BytesIO) -> BytesIOView:
        """Return a zero-copy view of the payload bytes."""
        return buffer.view(pos=self.start, size=self.size)

    def write_to(
        self,
        data: bytes | bytearray | memoryview,
        *,
        buffer: BytesIO | None = None,
    ) -> BytesIO:
        """Write header + payload into a buffer and return that buffer."""
        if buffer is None:
            buffer = BytesIO()

        payload = memoryview(data)
        if len(payload) != self.size:
            raise ValueError(
                f"Payload size mismatch: header.size={self.size}, actual={len(payload)}"
            )

        metadata_blob = encode_metadata(self.metadata)
        if len(metadata_blob) != self.meta_size:
            raise ValueError(
                f"Metadata size mismatch: header.meta_size={self.meta_size}, "
                f"actual={len(metadata_blob)}"
            )

        fixed = bytearray(HEADER_SIZE)
        fixed[0:2] = self.tag.to_bytes(2, "big")
        fixed[2:4] = self.codec.to_bytes(2, "big")
        fixed[4:8] = self.size.to_bytes(4, "big")
        fixed[8:12] = self.meta_size.to_bytes(4, "big")

        buffer.write_bytes(fixed)
        if metadata_blob:
            buffer.write_bytes(metadata_blob)
        buffer.write_bytes(payload)
        return buffer


"""Module dependency and pip index inspection utilities."""

# modules.py
from __future__ import annotations

import dataclasses as dc
import json
import os
import re
import shlex
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple, Iterable, MutableMapping

__all__ = [
    "PipIndexSettings",
    "get_pip_index_settings",
]


DEFAULT_PIP_INDEX_SETTINGS = None
_REQ_NAME_RE = re.compile(r"^\s*([A-Za-z0-9][A-Za-z0-9._-]*)")


def _run_pip(*args: str) -> Tuple[int, str, str]:
    """Run pip with arguments and return (returncode, stdout, stderr).

    Args:
        *args: Pip arguments.

    Returns:
        Tuple of (returncode, stdout, stderr).
    """
    p = subprocess.run(
        [sys.executable, "-m", "pip", *args],
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
    )
    return p.returncode, p.stdout.strip(), p.stderr.strip()


@dc.dataclass(frozen=True)
class PipIndexSettings:
    """Resolved pip index configuration from env and config sources."""
    index_url: Optional[str] = None
    extra_index_urls: List[str] = dc.field(default_factory=list)
    sources: Dict[str, Dict[str, Any]] = dc.field(default_factory=dict)  # {"env": {...}, "config": {...}}

    @classmethod
    def current(cls):
        """Return the cached default pip index settings.

        Returns:
            Default PipIndexSettings instance.
        """
        global DEFAULT_PIP_INDEX_SETTINGS

        if DEFAULT_PIP_INDEX_SETTINGS is None:
            DEFAULT_PIP_INDEX_SETTINGS = get_pip_index_settings()

        return DEFAULT_PIP_INDEX_SETTINGS

    @property
    def extra_index_url(self):
        """Return extra index URLs as a space-separated string.

        Returns:
            Space-separated extra index URLs or None.
        """
        if self.extra_index_urls:
            return " ".join(self.extra_index_urls)
        return None

    def setenv(
        self,
        env: MutableMapping[str, str] | None = None,
        *,
        keys: Iterable[str] | None = None
    ):
        keys = keys or ["UV_EXTRA_INDEX_URL", "PIP_EXTRA_INDEX_URL"]

        url = self.extra_index_url

        if env is None:
            env = os.environ

        for key in keys:
            env[key] = url

    def as_dict(self) -> dict:
        """Return a dict representation of the settings.

        Returns:
            Dict representation of settings.
        """
        return dc.asdict(self)


def get_pip_index_settings() -> PipIndexSettings:
    """
    Inspect pip settings:
      - env (PIP_INDEX_URL / PIP_EXTRA_INDEX_URL)
      - pip config (merged view from `pip config list`)

    Precedence:
      env overrides config.
    """
    sources: Dict[str, Dict[str, Any]] = {"env": {}, "config": {}}

    env_index = os.environ.get("PIP_INDEX_URL")
    env_extra = os.environ.get("PIP_EXTRA_INDEX_URL")

    if env_index:
        sources["env"]["PIP_INDEX_URL"] = env_index
    if env_extra:
        sources["env"]["PIP_EXTRA_INDEX_URL"] = env_extra

    env_extra_urls: List[str] = shlex.split(env_extra) if env_extra else []

    # Read pip config (best-effort)
    rc, out, _err = _run_pip("config", "list", "--format=json")
    config_index_url: Optional[str] = None
    config_extra_raw: List[Any] = []

    if rc == 0 and out:
        cfg = json.loads(out)
        for k, v in cfg.items():
            lk = k.lower()
            if lk.endswith("index-url"):
                sources["config"][k] = v
                if lk.endswith("index-url") and not lk.endswith("extra-index-url"):
                    config_index_url = str(v)
                elif lk.endswith("extra-index-url"):
                    if isinstance(v, list):
                        config_extra_raw.extend(v)
                    else:
                        config_extra_raw.append(v)
    else:
        rc2, out2, _ = _run_pip("config", "list")
        if rc2 == 0 and out2:
            for line in out2.splitlines():
                if "=" not in line:
                    continue
                k, v = [x.strip() for x in line.split("=", 1)]
                lk = k.lower()
                if lk.endswith("extra-index-url"):
                    sources["config"][k] = v
                    config_extra_raw.append(v)
                elif lk.endswith("index-url") and not lk.endswith("extra-index-url"):
                    sources["config"][k] = v
                    config_index_url = v

    # Apply precedence
    index_url = env_index or config_index_url

    # extras: if env is set, it replaces config (pip behavior)
    if env_extra_urls:
        candidates = list(env_extra_urls)  # already tokenized; do NOT split again
    else:
        # config entries might contain multiple URLs in a single string => split them
        candidates: List[str] = []
        for item in config_extra_raw:
            if item is None:
                continue
            candidates.extend(shlex.split(str(item)))

    # Dedup preserving order
    seen = set()
    extra_index_urls: List[str] = []
    for u in candidates:
        if u not in seen:
            seen.add(u)
            extra_index_urls.append(u)

    return PipIndexSettings(index_url=index_url, extra_index_urls=extra_index_urls, sources=sources)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

__all__ = ["Tags"]

if TYPE_CHECKING:
    from yggdrasil.pickle.ser.serialized import Serialized


class Tags:
    """
    Stable wire tags for serialized payload kinds.

    These integers are part of the binary protocol and must remain stable.

    Layout by reserved ranges
    -------------------------
    0..99    : primitive / logical scalars
    100..199 : collections
    200..299 : system / complex / runtime objects  (logging: 216–219)
    300..399 : framework-specific internal objects
    400..499 : pyarrow
    500..599 : pandas
    600..699 : polars
    700..799 : pyspark
    """

    # ------------------------------------------------------------------
    # runtime registries
    # ------------------------------------------------------------------

    CLASSES: ClassVar[dict[int, type["Serialized[object]"]]] = {}
    TYPES: ClassVar[dict[type, type["Serialized[object]"]]] = {}

    TAG_TO_NAME: ClassVar[dict[int, str]]

    # ------------------------------------------------------------------
    # category labels
    # ------------------------------------------------------------------

    CATEGORY_UNKNOWN: ClassVar[str] = "unknown"
    CATEGORY_PRIMITIVE: ClassVar[str] = "primitive"
    CATEGORY_COLLECTION: ClassVar[str] = "collection"
    CATEGORY_SYSTEM: ClassVar[str] = "system"
    CATEGORY_FRAMEWORK: ClassVar[str] = "framework"
    CATEGORY_ARROW: ClassVar[str] = "arrow"
    CATEGORY_PANDAS: ClassVar[str] = "pandas"
    CATEGORY_POLARS: ClassVar[str] = "polars"
    CATEGORY_PYSPARK: ClassVar[str] = "pyspark"

    # ------------------------------------------------------------------
    # category ranges
    # ------------------------------------------------------------------

    PRIMITIVE_BASE: ClassVar[int] = 0
    COLLECTION_BASE: ClassVar[int] = 100
    SYSTEM_BASE: ClassVar[int] = 200
    FRAMEWORK_BASE: ClassVar[int] = 300
    ARROW_BASE: ClassVar[int] = 400
    PANDAS_BASE: ClassVar[int] = 500
    POLARS_BASE: ClassVar[int] = 600
    PYSPARK_BASE: ClassVar[int] = 700
    CATEGORY_SIZE: ClassVar[int] = 100

    # ------------------------------------------------------------------
    # primitives / logical scalars
    # ------------------------------------------------------------------

    NONE: int = 0
    UTF8_STRING: int = 1
    LATIN1_STRING: int = 2
    BYTES: int = 3
    UINT8: int = 4
    INT8: int = 5
    UINT16: int = 6
    INT16: int = 7
    UINT32: int = 8
    INT32: int = 9
    UINT64: int = 10
    INT64: int = 11
    FLOAT16: int = 12
    FLOAT32: int = 13
    FLOAT64: int = 14
    BOOL: int = 15
    DECIMAL: int = 16
    DATETIME: int = 17
    DATE: int = 18
    TIME: int = 19
    TIMEDELTA: int = 20
    TIMEZONE: int = 21
    UUID: int = 22
    COMPLEX: int = 23
    IPADDRESS: int = 24

    # ------------------------------------------------------------------
    # collections
    # ------------------------------------------------------------------

    ARRAY: int = 100
    LIST: int = 101
    MAPPING: int = 102
    TUPLE: int = 103
    SET: int = 104
    GENERATOR: int = 105
    ITERATOR: int = 106
    FROZENSET: int = 107
    DEQUE: int = 108

    LARGE_ARRAY: int = 160
    LARGE_LIST: int = 161
    LARGE_MAPPING: int = 162
    LARGE_TUPLE: int = 163
    LARGE_SET: int = 164
    LARGE_GENERATOR: int = 165
    LARGE_ITERATOR: int = 166
    LARGE_FROZENSET: int = 167
    LARGE_DEQUE: int = 168

    # ------------------------------------------------------------------
    # system / complex
    # ------------------------------------------------------------------

    MODULE: int = 200
    CLASS: int = 201
    FUNCTION: int = 202
    BASE_EXCEPTION: int = 203
    PATH: int = 204
    IO: int = 205
    PICKLE: int = 206
    DILL: int = 207
    CLOUDPICKLE: int = 208
    DATACLASS: int = 209
    IO_BINARY = 210
    IO_TEXT = 211
    IO_BYTES_BUFFER = 212
    IO_STRING_BUFFER = 213
    URL: int = 214
    METHOD: int = 215
    LOGGING_LOGGER: int = 216
    LOGGING_HANDLER: int = 217
    LOGGING_FORMATTER: int = 218
    LOGGING_LOG_RECORD: int = 219

    # ------------------------------------------------------------------
    # arrow
    # ------------------------------------------------------------------

    ARROW_TABLE: int = 400
    ARROW_RECORD_BATCH: int = 401
    ARROW_STREAM: int = 402
    ARROW_DATASET: int = 403
    ARROW_SCHEMA: int = 404
    ARROW_FIELD: int = 405
    ARROW_DATA_TYPE: int = 406
    ARROW_ARRAY: int = 407
    ARROW_CHUNKED_ARRAY: int = 408
    ARROW_SCALAR: int = 409
    ARROW_TENSOR: int = 410

    # ------------------------------------------------------------------
    # pandas
    # ------------------------------------------------------------------

    PANDAS_DATAFRAME: int = 500
    PANDAS_SERIES: int = 501
    PANDAS_INDEX: int = 502

    # ------------------------------------------------------------------
    # polars
    # ------------------------------------------------------------------

    POLARS_DATAFRAME: int = 600
    POLARS_SERIES: int = 601
    POLARS_LAZYFRAME: int = 602
    POLARS_EXPR: int = 603
    POLARS_SCHEMA: int = 604
    POLARS_DATATYPE: int = 605

    # ------------------------------------------------------------------
    # pyspark
    # ------------------------------------------------------------------

    PYSPARK_DATAFRAME: int = 700
    PYSPARK_ROW: int = 701
    PYSPARK_SCHEMA: int = 702
    PYSPARK_DATATYPE: int = 703
    PYSPARK_COLUMN: int = 704
    PYSPARK_RDD: int = 705
    PYSPARK_SESSION: int = 706

    # ------------------------------------------------------------------
    # category helpers
    # ------------------------------------------------------------------

    @classmethod
    def _category_id(cls, tag: int) -> int:
        return -1 if tag < 0 else tag // cls.CATEGORY_SIZE

    @classmethod
    def get_category(cls, tag: int) -> str:
        cid = cls._category_id(tag)

        if cid == 0:
            return cls.CATEGORY_PRIMITIVE
        if cid == 1:
            return cls.CATEGORY_COLLECTION
        if cid == 2:
            return cls.CATEGORY_SYSTEM
        if cid == 3:
            return cls.CATEGORY_FRAMEWORK
        if cid == 4:
            return cls.CATEGORY_ARROW
        if cid == 5:
            return cls.CATEGORY_PANDAS
        if cid == 6:
            return cls.CATEGORY_POLARS
        if cid == 7:
            return cls.CATEGORY_PYSPARK
        return cls.CATEGORY_UNKNOWN

    @classmethod
    def is_primitive(cls, tag: int) -> bool:
        return 0 <= tag < 100

    @classmethod
    def is_collection(cls, tag: int) -> bool:
        return 100 <= tag < 200

    @classmethod
    def is_system(cls, tag: int) -> bool:
        return 200 <= tag < 300

    @classmethod
    def is_framework(cls, tag: int) -> bool:
        return 300 <= tag < 400

    @classmethod
    def is_arrow(cls, tag: int) -> bool:
        return 400 <= tag < 500

    @classmethod
    def is_pandas(cls, tag: int) -> bool:
        return 500 <= tag < 600

    @classmethod
    def is_polars(cls, tag: int) -> bool:
        return 600 <= tag < 700

    @classmethod
    def is_pyspark(cls, tag: int) -> bool:
        return 700 <= tag < 800

    # ------------------------------------------------------------------
    # lookup helpers
    # ------------------------------------------------------------------

    @classmethod
    def get_name(cls, tag: int) -> str | None:
        return cls.TAG_TO_NAME.get(tag)

    @classmethod
    def is_known(cls, tag: int) -> bool:
        return tag in cls.TAG_TO_NAME or tag in cls.CLASSES

    # ------------------------------------------------------------------
    # lazy import routing
    # ------------------------------------------------------------------

    @classmethod
    def _ensure_category_imported(cls, tag: int) -> None:
        """
        Import serializer modules lazily by tag category.

        This keeps import graphs lighter and reduces circular import drama.
        """
        cid = cls._category_id(tag)

        if cid == 0:
            from yggdrasil.pickle.ser.primitives import PrimitiveSerialized  # noqa: F401
            from yggdrasil.pickle.ser.logicals import LogicalSerialized  # noqa: F401
        elif cid == 1:
            from yggdrasil.pickle.ser.collections import CollectionSerialized  # noqa: F401
        elif cid == 2:
            from yggdrasil.pickle.ser.complexs import ComplexSerialized  # noqa: F401
            from yggdrasil.pickle.ser.ios import IOSerialized  # noqa: F401
            from yggdrasil.pickle.ser.pickles import PickleSerialized  # noqa: F401
            from yggdrasil.pickle.ser.logicals import PathSerialized  # noqa: F401
            from yggdrasil.pickle.ser.logging import LoggingSerialized  # noqa: F401
        elif cid == 4:
            from yggdrasil.pickle.ser.pyarrow import ArrowSerialized  # noqa: F401
        elif cid == 5:
            from yggdrasil.pickle.ser.pandas import PandasSerialized  # noqa: F401
        elif cid == 6:
            from yggdrasil.pickle.ser.polars import PolarsSerialized  # noqa: F401
        elif cid == 7:
            from yggdrasil.pickle.ser.pyspark import PySparkSerialized  # noqa: F401

    # ------------------------------------------------------------------
    # class resolution
    # ------------------------------------------------------------------

    @classmethod
    def get_class(cls, tag: int) -> type["Serialized[object]"]:
        """
        Resolve the concrete Serialized subclass registered for a wire tag.
        """
        existing = cls.CLASSES.get(tag)
        if existing is not None:
            return existing

        cls._ensure_category_imported(tag)
        existing = cls.CLASSES.get(tag)

        if existing is None:
            raise NotImplementedError(
                f"Tag {tag} ({cls.get_name(tag) or 'UNKNOWN'}) is not registered "
                f"with a Serialized class"
            )
        return existing

    @classmethod
    def get_class_from_type(cls, pytype: type) -> type["Serialized[object]"] | None:
        """
        Resolve the serializer class for a Python runtime type.

        Fast path:
            direct TYPES lookup

        Slow path:
            force-load known categories, then retry
        """
        existing = cls.TYPES.get(pytype)
        if existing is not None:
            return existing

        if not cls.TYPES:
            for cid in (0, 1, 2, 4):
                cls._ensure_category_imported(cid * cls.CATEGORY_SIZE)

        return cls.TYPES.get(pytype)

    @classmethod
    def resolve_class(cls, tag: int) -> type["Serialized[object]"]:
        return cls.get_class(tag)

    # ------------------------------------------------------------------
    # registration
    # ------------------------------------------------------------------

    @classmethod
    def register_class(
        cls,
        serialized_cls: type["Serialized[object]"],
        *,
        tag: int | None = None,
        pytype: type | None = None,
    ) -> None:
        """
        Register a Serialized subclass by wire tag and/or Python type.
        """
        if tag is None:
            tag = getattr(serialized_cls, "TAG", None)

        if isinstance(tag, int) and tag >= 0:
            cls.CLASSES[tag] = serialized_cls

        if pytype is not None:
            existing = cls.TYPES.get(pytype)

            if existing is None:
                cls.TYPES[pytype] = serialized_cls


def _build_tag_to_name() -> dict[int, str]:
    out: dict[int, str] = {}
    for name, value in vars(Tags).items():
        if name.startswith("_"):
            continue
        if not name.isupper():
            continue
        if isinstance(value, int):
            out[value] = name
    return out


Tags.TAG_TO_NAME = _build_tag_to_name()
Tags.TAG_TO_NAME = _build_tag_to_name()



"""Base options dataclass for :class:`~yggdrasil.io.buffer.media_io.MediaIO`.

:class:`MediaOptions` collects every parameter that can influence how an
Arrow table is read from / written to a :class:`~yggdrasil.io.buffer.BytesIO`
buffer.  Format-specific subclasses (:class:`ParquetOptions`,
:class:`JsonOptions`, …) extend it with codec-level knobs, but the
fields defined here are shared across *all* formats:

.. list-table::
   :header-rows: 1
   :widths: 20 10 70

   * - Field
     - Default
     - Description
   * - ``columns``
     - ``None``
     - Column names to read (``None`` = all).
   * - ``use_threads``
     - ``True``
     - Enable multi-threaded reads / writes.
   * - ``ignore_empty``
     - ``True``
     - Silently return an empty table when the buffer is empty.
   * - ``lazy``
     - ``False``
     - Return a lazy / streaming representation when the format supports it.
   * - ``raise_error``
     - ``True``
     - Raise on read/write errors instead of returning a sentinel.
   * - ``mode``
     - ``AUTO``
     - :class:`~yggdrasil.io.enums.SaveMode` governing the write strategy.
   * - ``match_by``
     - ``None``
     - Column names forming the composite key for ``UPSERT`` mode.

Save modes
----------
``AUTO`` / ``OVERWRITE``
    Replace the buffer contents entirely.

``APPEND``
    Concatenate new data after the existing data.

``UPSERT``
    Merge by ``match_by`` columns — new rows replace matching old rows,
    unmatched old rows survive, unmatched new rows are appended.

``IGNORE``
    Skip the write when the buffer already contains data.

``ERROR_IF_EXISTS``
    Raise :exc:`IOError` when the buffer already contains data.
"""
from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Optional, Sequence, Any

from ..enums.save_mode import SaveMode

__all__ = [
    "MediaOptions",
]

_MISSING = object()

_ALLOWED_COMPRESSION = {"auto", "none", "off", "zstd", "snappy", "gzip", "lz4"}


@dataclass(slots=True)
class MediaOptions:
    """Base options shared by all :class:`MediaIO` subclasses.

    Parameters
    ----------
    columns:
        Column names to read.  ``None`` means all columns.
    use_threads:
        Enable multi-threaded reading / writing.
    ignore_empty:
        When ``True``, an empty buffer returns an empty table instead of
        raising.
    lazy:
        When ``True``, return a lazy / streaming representation if the
        format supports it (e.g. a Polars ``LazyFrame``).
    raise_error:
        When ``True``, raise on read/write errors.
    mode:
        :class:`~yggdrasil.io.enums.SaveMode` governing the write
        strategy.  See the module docstring for semantics.
    match_by:
        Column names forming the composite deduplication key for
        ``UPSERT`` mode.  Ignored by other modes.
    """

    # global read properties
    columns: Optional[Sequence[str]] = None
    use_threads: bool = True
    ignore_empty: bool = True
    lazy: bool = False
    raise_error: bool = True

    # global write properties
    mode: SaveMode = SaveMode.AUTO
    match_by: Sequence[str] | None = None

    @classmethod
    def check_parameters(
        cls,
        options: Optional["MediaOptions"] = None,
        *,
        mode: SaveMode | str | None | Any = _MISSING,
        match_by: Sequence[str] | str | None | Any = _MISSING,
        columns: Optional[Sequence[str]] = _MISSING,
        use_threads: bool = _MISSING,
        ignore_empty: bool = _MISSING,
        lazy: bool = _MISSING,
        raise_error: bool = _MISSING,
        **kwargs,
    ) -> "MediaOptions":
        """Merge and validate parameters into an options instance.

        * Only explicitly supplied kwargs override existing values.
        * Unknown kwargs raise :exc:`TypeError`.
        * ``mode`` is normalised via :meth:`SaveMode.parse`.
        * ``match_by`` is normalised to ``tuple[str, …] | None``.
        * ``columns`` is normalised to ``list[str] | None``.

        Parameters
        ----------
        options:
            An existing instance to update, or ``None`` for a fresh default.
        mode, match_by, columns, use_threads, ignore_empty, lazy, raise_error:
            Overrides for the corresponding fields.
        **kwargs:
            Subclass-specific overrides (e.g. ``compression``).

        Returns
        -------
        MediaOptions
            The validated, merged instance (mutated in-place when *options*
            is an instance of *cls*; a new object otherwise).

        Raises
        ------
        TypeError
            On unknown kwargs, bad types for ``columns``, ``match_by``,
            boolean fields, or ``compression``.
        ValueError
            On invalid ``compression`` or ``skip_rows`` values.
        """
        if options is None:
            out = cls()
        else:
            if not isinstance(options, MediaOptions):
                raise TypeError(f"options must be a MediaOptions instance or None, got {type(options)!r}")

            if isinstance(options, cls):
                out = options
            else:
                # upcast into cls
                out = cls()
                for f in fields(cls):
                    if hasattr(options, f.name):
                        setattr(out, f.name, getattr(options, f.name))

        allowed = {f.name for f in fields(cls)}

        updates: dict[str, Any] = {}

        if mode is not _MISSING:
            updates["mode"] = mode
        if match_by is not _MISSING:
            updates["match_by"] = match_by
        if columns is not _MISSING:
            updates["columns"] = columns
        if use_threads is not _MISSING:
            updates["use_threads"] = use_threads
        if ignore_empty is not _MISSING:
            updates["ignore_empty"] = ignore_empty
        if lazy is not _MISSING:
            updates["lazy"] = lazy
        if raise_error is not _MISSING:
            updates["raise_error"] = raise_error

        updates.update(kwargs)

        unknown = [k for k in updates.keys() if k not in allowed]
        if unknown:
            raise TypeError(
                f"{cls.__name__}.check_parameters got unexpected parameter(s): {', '.join(sorted(unknown))}"
            )

        # ---- validations / normalizations ----

        if "columns" in updates:
            cols = updates["columns"]
            if cols is None:
                pass
            elif isinstance(cols, (str, bytes)):
                raise TypeError("columns must be a sequence of strings, not a single string/bytes")
            else:
                try:
                    cols_list = list(cols)
                except TypeError as e:
                    raise TypeError("columns must be a sequence of strings") from e
                if not all(isinstance(c, str) for c in cols_list):
                    bad = [type(c).__name__ for c in cols_list if not isinstance(c, str)]
                    raise TypeError(f"columns must contain only str, found: {bad[:3]}")
                updates["columns"] = list(cols_list)

        for b in ("use_threads", "ignore_empty", "lazy", "raise_error"):
            if b in updates and not isinstance(updates[b], bool):
                raise TypeError(f"{b} must be bool, got {type(updates[b]).__name__}")

        if "mode" in updates:
            updates["mode"] = SaveMode.parse(updates["mode"], default=SaveMode.AUTO)

        if "match_by" in updates:
            mb = updates["match_by"]
            if mb is None:
                pass
            elif isinstance(mb, (str, bytes)):
                if isinstance(mb, bytes):
                    raise TypeError("match_by must be str or a sequence of str, not bytes")
                updates["match_by"] = (mb,)
            else:
                try:
                    mb_list = list(mb)
                except TypeError as e:
                    raise TypeError("match_by must be a string or a sequence of strings") from e
                if not mb_list:
                    updates["match_by"] = None
                else:
                    if not all(isinstance(c, str) for c in mb_list):
                        bad = [type(c).__name__ for c in mb_list if not isinstance(c, str)]
                        raise TypeError(f"match_by must contain only str, found: {bad[:3]}")
                    updates["match_by"] = tuple(mb_list)

        # common subclass validations
        if "compression" in updates:
            comp = updates["compression"]
            if comp is not None:
                if not isinstance(comp, str):
                    raise TypeError(f"compression must be str|None, got {type(comp).__name__}")
                if comp not in _ALLOWED_COMPRESSION:
                    raise ValueError(f"compression must be one of {sorted(_ALLOWED_COMPRESSION)}, got {comp!r}")

        if "compression_level" in updates:
            lvl = updates["compression_level"]
            if lvl is not None and not isinstance(lvl, int):
                raise TypeError(f"compression_level must be int|None, got {type(lvl).__name__}")

        if "skip_rows" in updates:
            sr = updates["skip_rows"]
            if not isinstance(sr, int) or sr < 0:
                raise ValueError("skip_rows must be a non-negative int")

        if "zip_compression" in updates:
            zc = updates["zip_compression"]
            if not isinstance(zc, int):
                raise TypeError(f"zip_compression must be int, got {type(zc).__name__}")

        # apply
        for k, v in updates.items():
            setattr(out, k, v)

        return out



from .common import ErrorResponse, StrictModel
from .databricks import DatabricksSQLRequest, DatabricksSQLResponse
from .python import (
    CommandResult,
    CreateEnvRequest,
    DataFrameColumn,
    DataFramePayload,
    DeleteEnvRequest,
    DeleteEnvResponse,
    EnvInfo,
    EnvListResponse,
    EnvRefRequest,
    EnvResponse,
    ExcelExecuteRequest,
    ExcelExecuteResponse,
    ExcelPrepareRequest,
    ExcelPrepareResponse,
    ExecuteCodeRequest,
    ExecutionResponse,
    ManifestData,
    MutationResponse,
    PackageRequest,
    RequirementItem,
    RequirementsResponse,
)
from .system import HealthResponse, SystemInfoResponse

__all__ = [
    "StrictModel",
    "ErrorResponse",
    "SystemInfoResponse",
    "HealthResponse",
    "EnvInfo",
    "CommandResult",
    "RequirementItem",
    "DataFrameColumn",
    "DataFramePayload",
    "EnvRefRequest",
    "CreateEnvRequest",
    "DeleteEnvRequest",
    "PackageRequest",
    "ExecuteCodeRequest",
    "ExcelExecuteRequest",
    "ExcelPrepareRequest",
    "EnvResponse",
    "EnvListResponse",
    "DeleteEnvResponse",
    "RequirementsResponse",
    "MutationResponse",
    "ExecutionResponse",
    "ExcelExecuteResponse",
    "ExcelPrepareResponse",
    "ManifestData",
    "DatabricksSQLRequest",
    "DatabricksSQLResponse",
]
